#! /usr/bin/env python

import openturns as ot
import math
import sys

ot.TESTPREAMBLE()
# ot.Log.Show(ot.Log.ALL)


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


print("----- multi-obj -----")
dim = 2

# multi-objective zdt1 problem
f = ot.SymbolicFunction(
    ["x1", "x2"], ["x1", "var g := 1.0 + 9.0 * (x1 + x2); g * (1.0 - sqrt(x1 / g))"]
)
bounds = ot.Interval([0.0] * 2, [1.0] * 2)
ineq = ot.SymbolicFunction(["x1", "x2"], ["x1-x2"])  # x1>x2

# Create the initial population
size = 100
dist = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
pop0 = dist.getSample(size)

multi_obj = ["nsga2", "moead", "mhaco", "nspso"]
if "moead_gen" in ot.Pagmo.GetAlgorithmNames():
    multi_obj.append("moead_gen")  # pagmo>=2.19

for name in multi_obj:
    for use_ineq in [False, True]:
        zdt1 = ot.OptimizationProblem(f)
        zdt1.setBounds(bounds)
        if use_ineq:
            zdt1.setInequalityConstraint(ineq)
        algo = ot.Pagmo(zdt1, name, pop0)
        algo.setMaximumIterationNumber(10)
        algo.setBlockSize(8)
        # algo.setProgressCallback(progress)
        # algo.setStopCallback(stop)
        algo.run()
        result = algo.getResult()
        x = result.getFinalPoints()
        if use_ineq:
            for x1, x2 in x:
                assert x1 + 1e-5 >= x2, f"ineq constraint not verified: {x1} >= {x2}"
        y = result.getFinalValues()
        fronts = result.getParetoFrontsIndices()
        assert len(fronts) > 0, "no pareto"
        print(name, len(fronts))
        assert (
            result.getCallsNumber() == (algo.getMaximumIterationNumber() + 1) * size
        ), f"wrong size: {result.getCallsNumber()}"

# rosenbrock for the other algorithms
print("----- mono-obj -----")
dim = 2
f = ot.SymbolicFunction(["x1", "x2"], ["1+100*(x2-x1^2)^2+(1-x1)^2"])
ineq = ot.SymbolicFunction(["x1", "x2"], ["-x2"])  # x2<0
bounds = ot.Interval([-1.5] * dim, [1.5] * dim)
pop0 = ot.JointDistribution([ot.Uniform(-1.5, 1.5)] * 2).getSample(80)

for name in ot.Pagmo.GetAlgorithmNames():
    if name in multi_obj:
        continue
    for use_ineq in [False, True]:
        problem = ot.OptimizationProblem(f)
        problem.setBounds(bounds)
        if use_ineq:
            problem.setInequalityConstraint(ineq)
        algo = ot.Pagmo(problem, name, pop0)
        algo.setMaximumIterationNumber(10)
        algo.setBlockSize(8)
        # algo.setProgressCallback(progress)
        # algo.setStopCallback(stop)
        algo.run()
        result = algo.getResult()
        x = result.getOptimalPoint()
        y = result.getOptimalValue()
        if use_ineq:
            assert x[1] < 1e-5, f"ineq constraint not verified: {x[1]} < 0"
        else:
            assert result.getFinalPoints().getSize() == pop0.getSize(), "no final pop"
        assert y[0] < 40.0, str(y)
        print(name, x, y)
        if use_ineq:
            assert x[1] < 1e-5, "verified constraint"

print("----- minlp -----")


def minlp_obj(x):
    obj = 0
    for i in range(3):
        obj += (
            (x[2 * i - 2] - 3) ** 2 / 1000.0
            - (x[2 * i - 2] - x[2 * i - 1])
            + math.exp(20.0 * (x[2 * i - 2] - x[2 * i - 1]))
        )
    return [obj]


def minlp_cstr(x):
    ce1 = 4 * (x[0] - x[1]) ** 2 + x[1] - x[2] ** 2 + x[2] - x[3] ** 2
    ce2 = (
        8 * x[1] * (x[1] ** 2 - x[0])
        - 2 * (1 - x[1])
        + 4 * (x[1] - x[2]) ** 2
        + x[0] ** 2
        + x[2]
        - x[3] ** 2
        + x[3]
        - x[4] ** 2
    )
    ce3 = (
        8 * x[2] * (x[2] ** 2 - x[1])
        - 2 * (1 - x[2])
        + 4 * (x[2] - x[3]) ** 2
        + x[1] ** 2
        - x[0]
        + x[3]
        - x[4] ** 2
        + x[0] ** 2
        + x[4]
        - x[5] ** 2
    )
    ce4 = (
        8 * x[3] * (x[3] ** 2 - x[2])
        - 2 * (1 - x[3])
        + 4 * (x[3] - x[4]) ** 2
        + x[2] ** 2
        - x[1]
        + x[4]
        - x[5] ** 2
        + x[1] ** 2
        + x[5]
        - x[0]
    )
    ci1 = (
        8 * x[4] * (x[4] ** 2 - x[3])
        - 2 * (1 - x[4])
        + 4 * (x[4] - x[5]) ** 2
        + x[3] ** 2
        - x[2]
        + x[5]
        + x[2] ** 2
        - x[1]
    )
    ci2 = -(
        8 * x[5] * (x[5] ** 2 - x[4])
        - 2 * (1 - x[5])
        + x[4] ** 2
        - x[3]
        + x[3] ** 2
        - x[4]
    )
    return [-ce1, -ce2, -ce3, -ce4, -ci1, -ci2]


f = ot.PythonFunction(6, 1, minlp_obj)
bounds = ot.Interval([-5.0] * 6, [5.0] * 6)
ineq = ot.PythonFunction(6, 6, minlp_cstr)
pop0 = ot.JointDistribution(
    [ot.Uniform(-5.0, 5.0)] * 4
    + [ot.FiniteDiscreteDistribution([[i - 5] for i in range(11)])] * 2
).getSample(100)
problem = ot.OptimizationProblem(f)
problem.setBounds(bounds)
problem.setInequalityConstraint(ineq)
problem.setVariablesType(
    [ot.OptimizationProblemImplementation.CONTINUOUS] * 4
    + [ot.OptimizationProblemImplementation.INTEGER] * 2
)
for name in ["gaco", "ihs", "sga"]:
    algo = ot.Pagmo(problem, name, pop0)
    algo.setMaximumIterationNumber(10)
    algo.setBlockSize(8)
    algo.run()
    result = algo.getResult()
    x = result.getOptimalPoint()
    y = result.getOptimalValue()
    print(name, x, y)
    if name != "ihs":
        assert y[0] < 200.0, str(y)
print("-" * 20)


# check internal reordering of integer components
def minlp_obj2(x):
    x1, x2, x3 = x
    if float(int(x1)) != x1:
        raise ValueError(f"x1 should be int got {x1}")
    if max(abs(x2), abs(x3)) > 4.5:
        raise ValueError(f"x2,x3 not in bounds got {x2},{x3}")
    return [x1 + 100 * (x3 - x2**2) ** 2 + (1 - x2) ** 2]


f = ot.PythonFunction(3, 1, minlp_obj2)
bounds = ot.Interval([-5, -4.5, -4.5], [5, 4.5, 4.5])
pop0 = ot.JointDistribution([ot.Poisson()] + [ot.Uniform(-4.5, 4.5)] * 2).getSample(100)
problem = ot.OptimizationProblem(f)
problem.setBounds(bounds)
problem.setVariablesType(
    [
        ot.OptimizationProblemImplementation.INTEGER,
        ot.OptimizationProblemImplementation.CONTINUOUS,
        ot.OptimizationProblemImplementation.CONTINUOUS,
    ]
)
algo = ot.Pagmo(problem, "gaco", pop0)
algo.run()
result = algo.getResult()
x = result.getOptimalPoint()
y = result.getOptimalValue()
print("gaco reorder", x, y)
assert abs(-5.0 - y[0]) < 2e-3, f"wrong value {y}"

# check we don't expose penalized values
f = ot.SymbolicFunction(
    ["x1", "x2"], ["x1", "var g := 1.0 + 9.0 * (x1 + x2); g * (1.0 - sqrt(x1 / g))"]
)
zdt1 = ot.OptimizationProblem(f)
bounds = ot.Interval([0.0] * 2, [5.0] * 2)
zdt1.setBounds(bounds)
ineq = ot.ComposedFunction(
    ot.SymbolicFunction(["y1", "y2"], ["2-y1", "2-y2"]), f
)  # y1,y2 <2
zdt1.setInequalityConstraint(ineq)
dist = ot.JointDistribution([ot.Uniform(0.0, 5.0)] * 2)
pop0 = dist.getSample(50)
algo = ot.Pagmo(zdt1, "nsga2", pop0)
algo.setMaximumIterationNumber(10)
algo.run()
result = algo.getResult()
x = result.getFinalPoints()
y = result.getFinalValues()
for i in range(y.getSize()):
    assert y[i, 1] < 100.0, "penalized y value"
fi0 = result.getParetoFrontsIndices()[0]
print("penalized", fi0)
assert len(fi0) <= 6, "pareto indices " + str(fi0)



"""
Fitting a distribution with customized maximum likelihood
=========================================================
"""

# %%
# Introduction
# ------------
#
# When we perform distribution fitting using the :class:`~openturns.MaximumLikelihoodFactory`
# class, the default optimization solver sometimes fails to
# maximize the likelihood.
# This might be because the optimization solver is not appropriate in
# this particular case, or because the bounds of the problem
# are not properly set.
# In this example, we configure the optimization solver by two different
# methods.
#
# - The first method sets a specific solver among the :class:`~openturns.NLopt` solvers.
#   Moreover, we customize the distribution parameters bounds.
#   This helps the solver to avoid the part of the domain where
#   the distribution cannot be built.
# - The second method sets the :class:`~openturns.ResourceMap` key
#   that is used by the :class:`~openturns.MaximumLikelihoodFactory` in
#   order to select the default solver.

import openturns as ot

# %%
# Simulating a sample
# -------------------

# %%
# In this example, we consider the :class:`~openturns.TruncatedNormal` distribution
# with default parameters and generate a large sample from it.
distribution = ot.TruncatedNormal()
print("Distribution=", distribution)
size = 10000
sample = distribution.getSample(size)

# %%
# We can use the :class:`~openturns.TruncatedNormalFactory` class.
factory = ot.TruncatedNormalFactory()
fittedDistribution = factory.build(sample)
print("Fitted distribution=", fittedDistribution)

# %%
# But this method sometimes fails for specific distributions, for example
# when the sample size is very small or when the distribution has
# specific properties (e.g., some parameter is known beforehand).
# In these cases, general-purpose methods cannot be used, and the
# methods presented below may be relevant.

# %%
# Defining the estimation problem
# -------------------------------

# %%
# The truncated Normal distribution has 4 parameters:
#
# - the mean and standard deviation parameters of the
#   underlying Normal distribution,
# - the minimum and maximum of the underlying truncated distribution.
#
# Estimating the parameters will be easier if two conditions are met:
#
# - the starting point (i.e., the initial values of the parameters)
#   of the optimization is compatible with the sample,
# - the bounds are compatible with the distribution.
#
# First, we define the starting point of the optimization,
# using statistics from the sample.

# %%
sample_min = sample.getMin()[0]
sample_max = sample.getMax()[0]
sample_mean = sample.computeMean()[0]
sigma_hat = sample.computeStandardDeviation()[0]
startingPoint = [sample_mean, sigma_hat, sample_min, sample_max]
print("startingPoint=", startingPoint)

# %%
# Secondly, we set the bounds of the parameters.
# We assume that the minimum and maximum cannot be lower than -1
# or greater than 1.
# The parameters have the following bounds:
#
# - :math:`\mu \in [-\infty, +\infty]` ;
# - :math:`\sigma \in [0, +\infty]` ;
# - :math:`a \in [-1, 1]` ;
# - :math:`b \in [-1, 1]`.
#
# But this may not give satisfactory results, because infinite parameters
# are impossible and a zero standard deviation is infeasible.
# Furthermore, we must have :math:`a < b`.
# We assume that we know that the mean parameter cannot be greater than 10
# and lower than -10.
# Similarly, the standard deviation cannot be greater than 10.
# Hence, we use the following bounds:
#
# - :math:`\mu \in [-10, +10]` ;
# - :math:`\sigma \in [\epsilon, +10]` ;
# - :math:`a \in [-1 - \epsilon, \bar{x} - \epsilon]` ;
# - :math:`b \in [\bar{x} + \epsilon, 1 + \epsilon]`
#
# where :math:`\bar{x}` is the sample mean and :math:`\epsilon`
# is a small parameter.

# %%
epsilon = 1.0e-3
bounds_lower = [-10.0, epsilon, -1.0 - epsilon, sample_mean + epsilon]
bounds_upper = [10.0, 10.0, sample_mean - epsilon, 1.0 + epsilon]
print("bounds_lower=", bounds_lower)
print("bounds_upper=", bounds_upper)

# %%
# The boolean `False` means unbounded and `True` means bounded.

# %%
finiteLowerBound = [True] * 4
finiteUpperBound = [True] * 4
interval = ot.Interval(bounds_lower, bounds_upper, finiteLowerBound, finiteUpperBound)

# %%
# Setting the solver and bounds
# -----------------------------

# %%
# The two methods that we suggest are based on the :class:`~openturns.MaximumLikelihoodFactory` class
# which uses the maximum likelihood estimator to estimate
# the parameters of the distribution depending on the sample.
# The first method is to use the :meth:`~openturns.MaximumLikelihoodFactory.setOptimizationAlgorithm`
# method.
# This sets the optimization algorithm as a replacement for the
# default optimization solver.


def printTruncatedNormalParameters(truncatedNormalDistribution):
    mu, sigma, a, b = truncatedNormalDistribution.getParameter()
    print(" ", truncatedNormalDistribution.getImplementation().getClassName())
    print("  mu= %.3f" % (mu))
    print("  sigma= %.4f" % (sigma))
    print("  a= %.2f" % (a))
    print("  b= %.2f" % (b))
    return None


# %%
factory = ot.MaximumLikelihoodFactory(distribution)
factory.setOptimizationBounds(interval)
distribution_MLE = factory.build(sample)
print("Fitted distribution with default solver:")
printTruncatedNormalParameters(distribution_MLE)
default_solver = factory.getOptimizationAlgorithm()
print("Default solver=", default_solver.getImplementation().getClassName())
solver = ot.NLopt("LN_COBYLA")
print("New solver=", solver.getClassName(), solver.getAlgorithmName())
solver.setStartingPoint(startingPoint)
factory.setOptimizationAlgorithm(solver)
print(
    "Updated solver=",
    factory.getOptimizationAlgorithm().getImplementation().getClassName(),
)
distribution_MLE = factory.build(sample)
print("Fitted distribution with new solver:")
printTruncatedNormalParameters(distribution_MLE)


# %%
# Using ResourceMap to set the solver
# -----------------------------------

# %%
# Another method is to use a special key of the :class:`~openturns.ResourceMap`,
# which defines the default solver used by :class:`~openturns.MaximumLikelihoodFactory`.
# In this case, we do not define the starting point.

# %%
ot.ResourceMap.SetAsString(
    "MaximumLikelihoodFactory-DefaultOptimizationAlgorithm", "LN_COBYLA"
)
ot.ResourceMap.SetAsUnsignedInteger(
    "MaximumLikelihoodFactory-MaximumCallsNumber", 100000
)
factory = ot.MaximumLikelihoodFactory(distribution)
factory.setOptimizationBounds(interval)
print(
    "Default solver=",
    factory.getOptimizationAlgorithm().getImplementation().getClassName(),
)
distribution_MLE = factory.build(sample)
print("Fitted distribution:")
printTruncatedNormalParameters(distribution_MLE)

# %%
# Fitting a LogNormal with zero location parameter
# ------------------------------------------------

# %%
# We now consider the :class:`~openturns.LogNormal` distribution
# and present two different topics.
#
# - We show how to estimate the parameters of a LogNormal distribution
#   which has a zero location parameter.
# - We compute the list of solvers which are compatible with a specific
#   MLE problem.

# %%
# We generate a large sample from the standard LogNormal distribution.

# %%
standardLogNormal = ot.LogNormal()
sample = standardLogNormal.getSample(1000)

# %%
# If we estimate the parameters using :class:`~openturns.LogNormalFactory`,
# the algorithm does not know that the true value of the location parameter
# :math:`\gamma` is zero.
# Hence, the estimator may not necessarily lead to an estimated parameter
# exactly equal to zero.
# In this example, we assume that we know that this parameter is zero.
# In this case, the simplest method is to use the
# :meth:`~openturns.MaximumLikelihoodFactory.setKnownParameter` method.
# Furthermore, we know that the absolute value of the mean of the underlying Normal
# distribution cannot be greater than 5.
# Finally, we know that the standard deviation of the underlying
# Normal distribution cannot be lower than, say, :math:`\epsilon = 10^{-4}`.

# %%
logNormalFactoryWithZeroLocation = ot.MaximumLikelihoodFactory(standardLogNormal)
logNormalFactoryWithZeroLocation.setKnownParameter([2], [0.0])
lowerBound = [-5.0, 1.0e-4]
upperBound = [5.0, -1.0]
finiteLowerBound = [True, True]
finiteUpperBound = [True, False]
bounds = ot.Interval(lowerBound, upperBound, finiteLowerBound, finiteUpperBound)
logNormalFactoryWithZeroLocation.setOptimizationBounds(bounds)
fittedDistribution = logNormalFactoryWithZeroLocation.build(sample)
print("   fittedDistribution = ", fittedDistribution)

# %%
# See the list of solvers compatible with a given MLE problem
# -----------------------------------------------------------

# %%
# In the next script, we perform a loop over the algorithms
# from the :class:`~openturns.NLopt` class and select the algorithms
# which can solve the problem.
# This is done using `try/except` statements.
# Moreover, some solvers pretend to solve the problem, but do
# not make any change to the parameters.
# In this case, we do not select them.

# %%
defaultParameter = standardLogNormal.getParameter()
defaultSolver = logNormalFactoryWithZeroLocation.getOptimizationAlgorithm()
startingPoint = defaultSolver.getStartingPoint()
print("Default solver = ", defaultSolver.getImplementation().getClassName())
availableSolverList = []
differenceThreshold = 1.0e-8
for algorithmName in ot.NLopt.GetAlgorithmNames():
    print("- ", algorithmName)
    solver = ot.NLopt(algorithmName)
    solver.setStartingPoint(startingPoint)
    logNormalFactoryWithZeroLocation.setOptimizationAlgorithm(solver)
    try:
        fittedDistribution = logNormalFactoryWithZeroLocation.build(sample)
        print("   fittedDistribution = ", fittedDistribution)
        parameter = fittedDistribution.getParameter()
        differenceOfParameters = parameter - defaultParameter
        if differenceOfParameters.norm() > differenceThreshold:
            # Consider only solvers which make a difference after optimization
            availableSolverList.append(algorithmName)
    except TypeError:
        print("   Fail : Impossible to build with ", algorithmName)

# %%
print("Available solvers = ", len(availableSolverList))
for name in availableSolverList:
    print("- ", name)

# %%
# Pick one and try it.

# %%
algorithmName = "LD_SLSQP"
solver = ot.NLopt(algorithmName)
print("   New solver = ", solver.getClassName(), solver.getAlgorithmName())
solver.setStartingPoint(startingPoint)
logNormalFactoryWithZeroLocation.setOptimizationAlgorithm(solver)
fittedDistribution = logNormalFactoryWithZeroLocation.build(sample)
print("   fittedDistribution = ", fittedDistribution)



"""
Get the asymptotic distribution of the estimators
=================================================
"""

# %%
# In this example we introduce the `buildEstimator` method to obtain the asymptotic distribution of the parameters of a fitted distribution obtained from a `Factory`.
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# The standard Normal
# -------------------
# The parameters of the standard Normal distribution are estimated by a method of moments.
# Thus the asymptotic parameters distribution is Normal and estimated by bootstrap on the initial data.
#
distribution = ot.Normal(0.0, 1.0)
sample = distribution.getSample(50)
estimated = ot.NormalFactory().build(sample)

# %%
# We take a look at the estimated parameters :
print(estimated.getParameter())

# %%
# The `buildEstimator` method gives the asymptotic parameters distribution.
#
fittedRes = ot.NormalFactory().buildEstimator(sample)
paramDist = fittedRes.getParameterDistribution()

# %%
# We draw the 2d-PDF of the parameters
graph = paramDist.drawPDF()
graph.setXTitle(r"$\mu$")
graph.setYTitle(r"$\sigma$")
graph.setTitle(r"Normal fitting : $(\mu, \sigma)$ iso-PDF")
view = otv.View(graph)


# %%
# We draw the mean parameter :math:`\mu` distribution
graph = paramDist.getMarginal(0).drawPDF()
graph.setTitle(r"Normal fitting : PDF of $\mu$")
graph.setXTitle(r"$\mu$")
graph.setLegends(["PDF"])
view = otv.View(graph)

# %%
# We draw the scale parameter :math:`\sigma` distribution
graph = paramDist.getMarginal(1).drawPDF()
graph.setTitle(r"Normal fitting : PDF of $\sigma$")
graph.setXTitle(r"$\sigma$")
graph.setLegends(["PDF"])
view = otv.View(graph)

# %%
# We observe on the two previous figures that the distribution is Normal and centered around
# the estimated value of the parameter.


# %%
# The Pareto distribution
# -----------------------
#
# We consider a Pareto distribution with a scale parameter :math:`\beta=1.0`, a shape parameter :math:`\alpha=1.0` and a location parameter :math:`\gamma = 0.0`.
# We generate a sample from this distribution and use a :class:`~openturns.ParetoFactory` to fit the sample.
# In that case the asymptotic parameters distribution is estimated by bootstrap on the initial
# data and kernel fitting (see :class:`~openturns.KernelSmoothing`).
#

# %%
distribution = ot.Pareto(1.0, 1.0, 0.0)
sample = distribution.getSample(50)
estimated = ot.ParetoFactory().build(sample)

# %%
# We take a look at the estimated parameters :
print(estimated.getParameter())

# %%
# The `buildEstimator` method gives the asymptotic parameters distribution.
#
fittedRes = ot.ParetoFactory().buildEstimator(sample)
paramDist = fittedRes.getParameterDistribution()

# %%
# We draw scale parameter :math:`\beta` distribution
graph = paramDist.getMarginal(0).drawPDF()
graph.setTitle(r"Pareto fitting : PDF of $\beta$")
graph.setXTitle(r"$\beta$")
graph.setLegends(["PDF"])
view = otv.View(graph)

# %%
# We draw the shape parameter :math:`\alpha` distribution
graph = paramDist.getMarginal(1).drawPDF()
graph.setTitle(r"Pareto fitting : PDF of $\alpha$")
graph.setXTitle(r"$\alpha$")
graph.setLegends(["PDF"])
view = otv.View(graph)

# %%
# We draw the location parameter :math:`\gamma` distribution
graph = paramDist.getMarginal(2).drawPDF()
graph.setTitle(r"Pareto fitting : PDF of $\gamma$")
graph.setXTitle(r"$\gamma$")
graph.setLegends(["PDF"])
view = otv.View(graph)

# %%
# Display all graphs
otv.View.ShowAll()



"""
Estimate correlation coefficients
=================================
"""

# %%
# In this example we are going to estimate the correlation between an output sample Y and the corresponding inputs using various estimators:
#
# - Pearson coefficients
# - Spearman coefficients
# - PCC: Partial Correlation Coefficients
# - PRCC: Partial Rank Correlation Coefficient
# - SRC: Standard Regression Coefficients
# - SRRC: Standard Rank Regression Coefficient

# %%
from openturns.usecases import ishigami_function
import openturns as ot
import openturns.viewer as otv


# %%
# To illustrate the usage of the method mentioned above, we define a set of `X` and `Y` data using the :ref:`Ishigami model <use-case-ishigami>`.
# This classical model is defined in a data class:
im = ishigami_function.IshigamiModel()

# %%
# Create X/Y data
# We get the input variables description :
input_names = im.distribution.getDescription()

size = 100
inputDesign = ot.SobolIndicesExperiment(im.distribution, size, True).generate()
outputDesign = im.model(inputDesign)

# %%
# Create a :class:`~openturns.CorrelationAnalysis` object to compute various estimates
# of the correlation between the inputs and the output.

corr_analysis = ot.CorrelationAnalysis(inputDesign, outputDesign)

# %%
# PCC coefficients
# ------------------

pcc_indices = corr_analysis.computePCC()
print(pcc_indices)

# %%
#

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    pcc_indices, input_names, "PCC coefficients"
)
view = otv.View(graph)

# %%
# PRCC coefficients
# --------------------

prcc_indices = corr_analysis.computePRCC()
print(prcc_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    prcc_indices, input_names, "PRCC coefficients"
)
view = otv.View(graph)

# %%
# SRC coefficients
# -------------------

src_indices = corr_analysis.computeSRC()
print(src_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    src_indices, input_names, "SRC coefficients"
)
view = otv.View(graph)

# %%
# Normalized squared SRC coefficients (coefficients are made to sum to 1) :

squared_src_indices = corr_analysis.computeSquaredSRC(True)
print(squared_src_indices)

# %%

graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    squared_src_indices, input_names, "Squared SRC coefficients"
)
view = otv.View(graph)


# %%
# SRRC coefficients
# --------------------

srrc_indices = corr_analysis.computeSRRC()
print(srrc_indices)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    srrc_indices, input_names, "SRRC coefficients"
)
view = otv.View(graph)

# %%
# Pearson coefficients
# -----------------------
# We compute here the Pearson :math:`\rho` coefficients.

pearson_correlation = corr_analysis.computeLinearCorrelation()
print(pearson_correlation)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    pearson_correlation, input_names, "Pearson correlation coefficients"
)
view = otv.View(graph)

# %%
# Spearman coefficients
# -----------------------
# We compute here the Spearman :math:`\rho_s` coefficients.

# %%
spearman_correlation = corr_analysis.computeSpearmanCorrelation()
print(spearman_correlation)

# %%
graph = ot.SobolIndicesAlgorithm.DrawCorrelationCoefficients(
    spearman_correlation, input_names, "Spearman correlation coefficients"
)
view = otv.View(graph)


# %%
# Display all the figures.
otv.View.ShowAll()



#! /usr/bin/env python

import sys
import os
import shutil
from tempfile import mkdtemp
import doctest

use_matplotlib = False
try:
    import matplotlib

    matplotlib.use("Agg")
    use_matplotlib = True
except Exception:
    pass

py_files = sys.argv[1:]

total_failure_count = 0
total_test_count = 0

# Temporary working directory for IO tests
work_dir = mkdtemp()
os.chdir(work_dir)

for py_file in py_files:
    if (not use_matplotlib) and (os.path.basename(py_file) == "viewer.py"):
        failure_count, test_count = 0, 0
    else:
        py_basename = os.path.splitext(os.path.basename(py_file))[0]
        module = __import__("openturns." + py_basename, fromlist=[py_basename])
        failure_count, test_count = doctest.testmod(
            module, verbose=False, optionflags=doctest.ELLIPSIS
        )

    total_failure_count += failure_count
    total_test_count += test_count

    print(("%s %5d tests failed" % ((py_basename + " ").ljust(60, "."), failure_count)))

print(("-" * 79))
print(("%s %5d tests failed" % ("TOTAL ".ljust(60, "."), total_failure_count)))

# Delete temporary working directory for IO tests
os.chdir("..")
shutil.rmtree(work_dir)

sys.exit(total_failure_count)



import openturns as ot
import openturns.viewer as otv

data = ot.FrankCopula().getSample(1000)
graph = ot.VisualTest.DrawUpperExtremalDependenceFunction(data)

otv.View(graph, figure_kw={"figsize": (4.5, 4.5)})



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Logistic(2.0, 2.5)
size = 10000
sample = distribution.getSample(size)
factory = ot.LogisticFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedLogistic = factory.buildAsLogistic(sample)
print("Logistic          =", distribution)
print("Estimated logistic=", estimatedLogistic)
estimatedLogistic = factory.buildAsLogistic()
print("Default logistic=", estimatedLogistic)
estimatedLogistic = factory.buildAsLogistic(distribution.getParameter())
print("Logistic from parameters=", estimatedLogistic)



"""
Export a metamodel
------------------
"""

# %%
# In this example we will see how to export a metamodel from the context it is created
# to another context where it will actually be used with the help of the pickle module.

# %%
import openturns as ot
from openturns.usecases import cantilever_beam
import pickle

# %%
# Load the cantilever beam use-case.
# We want to use the independent distribution to get meaningful Sobol' indices.
beam = cantilever_beam.CantileverBeam()
g = beam.model
distribution = beam.independentDistribution

# %%
# Generate a learning sample with Monte-Carlo simulation (or retrieve the design from experimental data).
N = 30  # size of the experimental design
X = distribution.getSample(N)
Y = g(X)

# %%
# Build a chaos metamodel
algo = ot.FunctionalChaosAlgorithm(X, Y, distribution)
algo.run()
metamodel = algo.getResult().getMetaModel()

# %%
# Save the metamodel into a `.pkl` file for later use
with open("metamodel.pkl", "wb") as f:
    pickle.dump(metamodel, f)

# %%
# Reload the metamodel in another context from the same `.pkl`
with open("metamodel.pkl", "rb") as f:
    metamodel = pickle.load(f)

# %%
# Reuse the loaded metamodel
x = [6.70455e10, 300.0, 2.55, 1.45385e-07]
y = metamodel(x)
print(y)



#! /usr/bin/env python

from openturns.testing import assert_almost_equal
from openturns.usecases import oscillator
import openturns as ot

"""
Test the Oscillator data class on a reference point.
"""
osc = oscillator.Oscillator()

assert_almost_equal(osc.model(osc.distribution.getMean()), [23.1897], 1e-2)

# Compute probability with FORM
model = osc.model
distribution = osc.distribution
vect = ot.RandomVector(distribution)
G = ot.CompositeRandomVector(model, vect)
event = ot.ThresholdEvent(G, ot.Less(), 0.0)
event.setName("failure")

optimAlgo = ot.Cobyla()
optimAlgo.setMaximumCallsNumber(1000)
optimAlgo.setMaximumAbsoluteError(1.0e-3)
optimAlgo.setMaximumRelativeError(1.0e-3)
optimAlgo.setMaximumResidualError(1.0e-3)
optimAlgo.setMaximumConstraintError(1.0e-3)
optimAlgo.setStartingPoint(distribution.getMean())

algo = ot.FORM(optimAlgo, event)
algo.run()
result = algo.getResult()

assert_almost_equal(result.getEventProbability(), 3.8906e-06, 1e-3)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(6)
ot.ResourceMap.SetAsUnsignedInteger(
    "LinearCombinationDistribution-DefaultMaxSize", 4000000
)


# 1) Analytical test ==> No FFT
coll = ot.DistributionCollection(3)
coll[0] = ot.Normal(0.0, 1.0)
coll[1] = ot.Uniform(2.0, 5.0)
coll[2] = ot.Uniform(2.0, 5.0)
weights = ot.Matrix([[1.0, 2.0, 4.0], [3.0, 4.0, 5.0], [6.0, 0.0, 1.0]])
distribution = ot.LinearCombinationDistribution(coll, weights)
print("distribution=", repr(distribution))
print("range = ", distribution.getRange())
print("mean = ", distribution.getMean())
print("cov = ", distribution.getCovariance())
print("sigma = ", distribution.getStandardDeviation())
N = 4
points = ot.Indices(3, N)
mean = distribution.getMean()
sigma = distribution.getStandardDeviation()
xMin = mean - 2.9 * sigma
xMax = mean + 2.9 * sigma
grid = ot.Sample()
result, grid = distribution.computePDF(xMin, xMax, points)
print("x;y;z;PDF")
for i in range(grid.getSize()):
    print("%.6g;%.6g;%.6g;%.6g" % (grid[i][0], grid[i][1], grid[i][2], result[i][0]))

# 2) 3D test using FFT
collection = ot.DistributionCollection(0)
mixture = ot.Mixture([ot.Normal(2.0, 1.0), ot.Normal(-2.0, 1.0)])
collection3D = ot.DistributionCollection(0)
collection3D.add(ot.Normal(0.0, 1.0))
collection3D.add(mixture)
collection3D.add(ot.Uniform(0, 1))
collection3D.add(ot.Uniform(0, 1))
# / Set weights
weights = ot.Matrix(
    [[1.0, -0.05, 1.0, -0.5], [0.5, 1.0, -0.05, 0.3], [-0.5, -0.1, 1.2, -0.8]]
)
# Defining LinearCombinationDistribution
dist_3D = ot.LinearCombinationDistribution(collection3D, weights)
mean = dist_3D.getMean()
sigma = dist_3D.getStandardDeviation()
print("distribution = ", repr(dist_3D))
print("range = ", dist_3D.getRange())
print("mean = ", dist_3D.getMean())
print("cov = ", dist_3D.getCovariance())
print("sigma = ", dist_3D.getStandardDeviation())
xMin = mean - 2.9 * sigma
xMax = mean + 2.9 * sigma
print("xMin = ", xMin)
print("xMax = ", xMax)
result, grid = dist_3D.computePDF(xMin, xMax, points)
for i in range(grid.getSize()):
    print("%.6g;%.6g;%.6g;%.6g" % (grid[i][0], grid[i][1], grid[i][2], result[i][0]))



#!/usr/bin/env python

import openturns as ot


def printResults(result, problemName):
    print(f"*** {problemName} completed:")
    print(f"      -- Optimal point = {result.getOptimalPoint()}")
    print(f"      -- Optimal value = {result.getOptimalValue()}")
    print(f"      -- Iteration number = {result.getIterationNumber()}")
    print(f"      -- Evaluation number = {result.getCallsNumber()}")
    print(f"      -- Absolute error = {result.getAbsoluteError():.6e}")
    print(f"      -- Relative error = {result.getRelativeError():.6e}")
    print(f"      -- Residual error = {result.getResidualError():.6e}")
    print(f"      -- Constraint error = {result.getConstraintError():.6e}")


# Define the problems based on Rastrigin function
rastrigin = ot.SymbolicFunction(
    ["x1", "x2"], ["20 + x1^2 - 10*cos(2*pi_*x1) + x2^2 - 10*cos(2*pi_*x2)"]
)

unboundedProblem = ot.OptimizationProblem(rastrigin)

notConstrainingBounds = ot.Interval([-5.0, -5.0], [3.0, 2.0])
notConstrainingBoundsProblem = ot.OptimizationProblem(
    rastrigin, ot.Function(), ot.Function(), notConstrainingBounds
)

constrainingBounds = ot.Interval([-1.0, -2.0], [5.0, -0.5])
constrainingBoundsProblem = ot.OptimizationProblem(
    rastrigin, ot.Function(), ot.Function(), constrainingBounds
)

boundedPref = [0.0, -1.0]
unboundedPref = [0.0, 0.0]

# GLOBAL ALGORITHM

# Non-contraining bounds Global
notConstrainingBoundsGlobal = ot.Dlib(notConstrainingBoundsProblem, "global")
notConstrainingBoundsGlobal.setStartingPoint([0.0] * 2)
notConstrainingBoundsGlobal.setMaximumCallsNumber(300)
notConstrainingBoundsGlobal.run()
printResults(notConstrainingBoundsGlobal.getResult(), "Non-constraining bounds Global")

# Contraining bounds Global
constrainingBoundsGlobal = ot.Dlib(constrainingBoundsProblem, "global")
constrainingBoundsGlobal.setMaximumCallsNumber(300)
constrainingBoundsGlobal.setStartingPoint([0.0] * 2)
constrainingBoundsGlobal.run()
printResults(constrainingBoundsGlobal.getResult(), "Constraining bounds Global")

assert (
    notConstrainingBoundsGlobal.getResult().getOptimalValue()[0] < 4.0
), "optimum not found"



#! /usr/bin/env python

from openturns.viewer import View
import openturns as ot

# Curve
graph = ot.Normal().drawCDF()
view = View(graph, pixelsize=(800, 600), plot_kw={"color": "blue"})
# view.save('curve1.png')
view.show()

# Contour
graph = ot.Normal([1, 2], [3, 5], ot.CorrelationMatrix(2)).drawPDF()
view = View(graph)
# view.save('curve2.png')
view.show()

# Histogram tests
normal = ot.Normal(1)
size = 100
sample = normal.getSample(size)
graph = ot.HistogramFactory().build(sample, 10).drawPDF()
view = View(graph)
# view.save('curve3.png')
view.show()

# QQPlot tests
size = 100
normal = ot.Normal(1)
sample = normal.getSample(size)
sample2 = ot.Gamma(3.0, 4.0, 0.0).getSample(size)
graph = ot.VisualTest.DrawQQplot(sample, sample2, 100)
view = View(graph)
# view.save('curve4.png')
view.ShowAll(block=True)

# Clouds tests
dimension = 2
R = ot.CorrelationMatrix(dimension)
R[0, 1] = 0.8
distribution = ot.Normal(ot.Point(dimension, 3.0), ot.Point(dimension, 2.0), R)
size = 100
sample1 = ot.Normal([3.0] * dimension, [2.0] * dimension, R).getSample(size)
sample2 = ot.Normal([2.0] * dimension, [3.0] * dimension, R).getSample(size // 2)
cloud1 = ot.Cloud(sample1, "blue", "fsquare", "Sample1 Cloud")
cloud2 = ot.Cloud(sample2, "red", "fcircle", "Sample2 Cloud")
graph = ot.Graph("two samples clouds", "x1", "x2")
graph.setLegendPosition("upper right")
graph.add(cloud1)
graph.add(cloud2)
view = View(graph)
# view.save('curve5.png')
view.show()

# Cloud, empty legend
graph = ot.Graph("two samples clouds", "x1", "x2")
graph.setLegendPosition("upper right")
cloud1 = ot.Cloud(sample1)
graph.add(cloud1)
view = View(graph)
view.show()

# Text
graph = ot.Graph("Annotated cloud", "x", "y")
distribution = ot.Normal(2)
size = 30
sample2D = distribution.getSample(size)
cloud = ot.Cloud(sample2D, "red", "fsquare", "Sample2D Cloud")
graph.add(cloud)

# Display extrema indices
x1 = [x[0] for x in sample2D[:, 0]]
x2 = [x[0] for x in sample2D[:, 1]]
idx = [0] * 4
idx[0] = x1.index(min(x1))
idx[1] = x1.index(max(x1))
idx[2] = x2.index(min(x2))
idx[3] = x2.index(max(x2))

labels = ot.Description(sample2D.getSize())
for i in range(4):
    labels[idx[i]] = str(idx[i])

position = ot.Description(sample2D.getSize(), "top")
position[idx[0]] = "right"
position[idx[1]] = "left"
position[idx[2]] = "top"
position[idx[3]] = "bottom"

text = ot.Text(sample2D, labels)
text.setColor("red")
text.setTextPositions(position)

graph.add(text)
view = View(graph)
view.show()

# CobWeb tests
size = 100
dim = 6
inputSample = ot.Normal(dim).getSample(size)
inputVar = list(["X" + str(i) for i in range(dim)])
expression = ""
for i in range(dim):
    if i > 0:
        expression += "+"
    expression += "cos(" + str(i + 1) + "*" + inputVar[i] + ")"
model = ot.SymbolicFunction(inputVar, [expression])
outputSample = model(inputSample)
graph = ot.VisualTest.DrawParallelCoordinates(
    inputSample, outputSample, 2.5, 3.0, "red", False
)
view = View(graph, legend_kw={"loc": "lower center"})
# view.save('curve6.png')
view.show()

# Staircase
distribution = ot.Poisson(10.0)
graph = distribution.drawCDF()
view = View(graph)
# view.save('curve7.png')
view.ShowAll(block=True)

# Pie
graph = ot.SobolIndicesAlgorithm.DrawImportanceFactors(
    [0.4, 0.3, 0.2, 0.1], ["a0", "a1", "a2", "a3"], "Zou"
)
view = View(graph)
# view.save('curve8.png')
view.show()

# Convergence graph curve
aCollection = []
aCollection.append(
    ot.LogNormalFactory().build(ot.LogNormalMuSigma()([300.0, 30.0, 0.0]))
)
aCollection.append(ot.Normal(75e3, 5e3))
myDistribution = ot.JointDistribution(aCollection)
vect = ot.RandomVector(myDistribution)
LimitState = ot.SymbolicFunction(("R", "F"), ("R-F/(pi_*100.0)",))
G = ot.CompositeRandomVector(LimitState, vect)
myEvent = ot.ThresholdEvent(G, ot.Less(), 0.0)
experiment = ot.MonteCarloExperiment()
myAlgo = ot.ProbabilitySimulationAlgorithm(myEvent, experiment)
myAlgo.setMaximumCoefficientOfVariation(0.05)
myAlgo.setMaximumOuterSampling(int(1e5))
myAlgo.run()
graph = myAlgo.drawProbabilityConvergence()
view = View(graph)
# view.save('curve10.png')
view.show()

# Polygon
size = 50
cursor = [0.0] * 2
data1 = ot.Sample(size, 2)  # polygon y = 2x for x in [-25]
data2 = ot.Sample(size, 2)  # polygon y = x*x for x in [-11]
for i in range(size):
    tmp = 7.0 * i / size + 2
    cursor[0] = tmp
    cursor[1] = 2 * tmp
    data1[i] = cursor
    tmp = 9.0 * i / size + 1
    cursor[0] = tmp
    cursor[1] = tmp * tmp
    data2[i] = cursor
graph = ot.Graph("Some polygons", "x1", "x2")
graph.setLegendPosition("upper right")
myPolygon1 = ot.Polygon(data1)
myPolygon1.setColor("blue")
myPolygon1.setLegend("polygon 1")
graph.add(myPolygon1)
myPolygon2 = ot.Polygon(data2)
myPolygon2.setColor("red")
myPolygon2.setLegend("polygon 2")
graph.add(myPolygon2)
view = View(graph)
# view.save('curve11.png')
view.ShowAll(block=True)

# PolygonArray
generator = ot.Normal(2)
size = 5
array = [None] * size
palette = ot.Drawable.BuildDefaultPalette(size)
palette[0] = "blue"
for i in range(size):
    vertices = generator.getSample(3)
    array[i] = ot.Polygon(vertices, palette[i], palette[size - i - 1])
graph = ot.Graph("An array of polygons", "x", "y")
graph.setLegendPosition("upper right")
parray = ot.PolygonArray(array)
parray.setLegend("array of polys")
graph.add(parray)
view = View(graph)
# view.save('curve12.png')

# BipartiteGraph
graph = ot.BipartiteGraph([[0, 1, 2], [0, 1, 2]]).draw()
view = View(graph)

# BarPlot: FORM reliability index marginal parameter sensitivity
f = ot.SymbolicFunction(["E", "F", "L", "I"], ["-F*L^3/(3*E*I)"])
dim = f.getInputDimension()
mean = [50.0, 1.0, 10.0, 5.0]
sigma = ot.Point(dim, 1.0)
R = ot.IdentityMatrix(dim)
distribution = ot.Normal(mean, sigma, R)
vect = ot.RandomVector(distribution)
output = ot.CompositeRandomVector(f, vect)
event = ot.ThresholdEvent(output, ot.Less(), -3.0)
solver = ot.Cobyla()
solver.setMaximumCallsNumber(400)
solver.setMaximumAbsoluteError(1.0e-10)
solver.setMaximumRelativeError(1.0e-10)
solver.setMaximumResidualError(1.0e-10)
solver.setMaximumConstraintError(1.0e-10)
solver.setStartingPoint(mean)
algo = ot.FORM(solver, event)
algo.run()
result = algo.getResult()
graph, _ = result.drawHasoferReliabilityIndexSensitivity()
view = View(graph)

# Optimization error history
opt_result = result.getOptimizationResult()
graph = opt_result.drawErrorHistory()
view = View(graph)
view.ShowAll(block=True)

# GridLayout
grid = ot.GridLayout(2, 3)
palette = ot.Drawable.BuildDefaultPalette(10)
for j in range(grid.getNbColumns()):
    alpha = 1.0 + j
    pdf_curve = ot.WeibullMin(1.0, alpha, 0.0).drawPDF()
    cdf_curve = ot.WeibullMin(1.0, alpha, 0.0).drawCDF()
    pdf_curve.setColors([palette[j]])
    cdf_curve.setColors([palette[j]])
    legends = [f"alpha={alpha}"]
    pdf_curve.setLegends(legends)
    cdf_curve.setLegends(legends)
    grid.setGraph(0, j, pdf_curve)
    grid.setGraph(1, j, cdf_curve)
view = View(grid)

# Square axes
graph = ot.ClaytonCopula(5.0).drawPDF()
view = View(graph, square_axes=True)

# Show axes as prescribed by getAxes()
graph = ot.Normal().drawPDF()
graph.setAxes(False)
view = View(graph)

# test _repr_png_
png = graph._repr_png_()
assert b"PNG" in png[:10]

# BuildDefaultPalette, BuildTableauPalette
ncurves = 5
graph = ot.Graph("BuildPalette", "X", "Y")
graph.setLegendPosition("upper right")
n = 20
x = ot.Sample([[i] for i in range(n)])
for i in range(ncurves):
    y = ot.Normal().getSample(n)
    curve = ot.Curve(x, y)
    curve.setLegend("Curve #%d" % (i))
    graph.add(curve)
palette = ot.Drawable.BuildDefaultPalette(ncurves)
graph.setColors(palette)
view = View(graph)
palette = ot.Drawable.BuildTableauPalette(ncurves)
graph.setColors(palette)
view = View(graph)

# mixed legend
f = ot.SymbolicFunction(["x", "y"], ["sin(x)*sin(y)"])
# use different numbers of points for x and y to check
# these numbers are handled properly by the viewer
graph = f.draw([-4.0] * 2, [4.0] * 2, [64, 32])
curve = ot.Curve([-4.0, 4.0], [1.0, 1.0], "curve")
curve.setColor("black")
curve.setLineStyle("dashed")
graph.add(curve)
cloud = ot.Cloud(ot.Normal(2).getSample(100), "cloud")
graph.add(cloud)
view = View(graph)

# contour bug
outputSample = ot.Uniform().getSample(49)
x = [[-3.58], [-2.38667], [-1.19333], [0], [1.19333], [2.38667], [3.58]]
y = [[-3.58], [-2.38667], [-1.19333], [0], [1.19333], [2.38667], [3.58]]
contour = ot.Contour(x, y, outputSample)
contour.setLevels([0.0])
contour.setLabels(["0.0"])
graph = ot.Graph("", "", "")
graph.add(contour)
view = View(graph)

view.ShowAll(block=True)



#! /usr/bin/env python

import sys
import openturns as ot

ot.TESTPREAMBLE()
ot.Log.Show(ot.Log.INFO)


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


dimension = 3

# gsobol
# meanTh = 1.0
# a = [0.0] * dimension
# inputVariables = [''] * dimension
# formula = '1.0'
# covTh = 1.0
# for i in range(dimension):
# a[i] = 0.5 * i
# covTh = covTh * (1.0 + 1.0 / (3.0 * (1.0 + a[i]) ** 2))
# inputVariables[i] = "xi" + str(i)
# formula += " * ((abs(4.0 * xi" + str(i) + " - 2.0) + " + str(a[i]) + ") / (1.0 + " + str(a[i]) + "))"
# covTh = covTh - 1.0
# model = ot.SymbolicFunction(inputVariables, [formula])
# distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * dimension)

# ishigami
formula = "sin(pi_*X1)+7*sin(pi_*X2)*sin(pi_*X2)+0.1*((pi_*X3)*(pi_*X3)*(pi_*X3)*(pi_*X3))*sin(pi_*X1)"
model = ot.SymbolicFunction(["X1", "X2", "X3"], [formula])
distribution = ot.JointDistribution([ot.Uniform(-1.0, 1.0)] * dimension)

estimator = ot.SaltelliSensitivityAlgorithm()
estimator.setUseAsymptoticDistribution(True)
algo = ot.SobolSimulationAlgorithm(distribution, model, estimator)
algo.setMaximumOuterSampling(250)
# size N of Sobol experiment at each iteration, total size is N*(d+2)
algo.setExperimentSize(10000)
algo.setBlockSize(97)
algo.setIndexQuantileLevel(0.05)
algo.setIndexQuantileEpsilon(1e-2)
# algo.setProgressCallback(progress)
# algo.setStopCallback(stop)

print("algo=", algo)

# Perform the simulation
algo.run()

# Stream out the result
result = algo.getResult()
print("result=", result)
print("FO=", result.getFirstOrderIndicesEstimate())
print("TO=", result.getTotalOrderIndicesEstimate())

foDist = result.getFirstOrderIndicesDistribution()
print(foDist)
toDist = result.getTotalOrderIndicesDistribution()
print(toDist)
convergenceGraph = algo.drawFirstOrderIndexConvergence()
convergenceGraph = algo.drawTotalOrderIndexConvergence()
graph = result.draw()
print(graph)


# from openturns.viewer import View
# View(convergenceGraph).ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)

size = 10000

# factory.build from distribution sample
distribution = ot.Gamma(0.2, 1.0, 1.0)
sample = distribution.getSample(size)
factory = ot.GammaFactory()
estimatedDistribution = factory.build(sample)
ott.assert_almost_equal(estimatedDistribution, distribution, 0.13, 0.0)

# factory.build from distribution (with other parameters) sample
distribution = ot.Gamma(2.3, 1.0, 1.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
ott.assert_almost_equal(estimatedDistribution, distribution, 0.07, 0.0)

# factory.build from different sample of the same distribution
distribution = ot.Gamma(2.3, 1.0, 1.0)
sample = distribution.getSample(size)
estimatedDistribution = factory.build(sample)
ott.assert_almost_equal(estimatedDistribution, distribution, 0.07, 0.0)

# factory.build default
defaultDistribution = ot.Gamma(1.0, 1.0, 0.0)
estimatedDistribution = factory.build()
ott.assert_almost_equal(estimatedDistribution, defaultDistribution, 0.0, 0.0)

# factory.build from parameters
estimatedDistribution = factory.build(distribution.getParameter())
ott.assert_almost_equal(estimatedDistribution, distribution, 0.0, 0.0)

# factory.buildAs from distribution sample
estimatedGamma = factory.buildAsGamma(sample)
ott.assert_almost_equal(estimatedGamma, distribution, 0.07, 0.0)

# factory.buildAs default
estimatedGamma = factory.buildAsGamma()
ott.assert_almost_equal(estimatedGamma, defaultDistribution, 0.0, 0.0)

# factory.buildAs from distribution parameters
estimatedGamma = factory.buildAsGamma(distribution.getParameter())
ott.assert_almost_equal(estimatedGamma, distribution, 0.0, 0.0)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


distribution = ot.Bernoulli(0.7)
size = 10000
sample = distribution.getSample(size)
factory = ot.BernoulliFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedBernoulli = factory.buildAsBernoulli(sample)
print("Bernoulli          =", distribution)
print("Estimated Bernoulli=", estimatedBernoulli)
estimatedBernoulli = factory.buildAsBernoulli()
print("Default Bernoulli=", estimatedBernoulli)
estimatedBernoulli = factory.buildAsBernoulli(distribution.getParameter())
print("Bernoulli from parameters=", estimatedBernoulli)
result = factory.buildEstimator(sample)
estimatedDistribution = result.getDistribution()
print("Estimated distribution            =", repr(estimatedDistribution))
parameterDistribution = result.getParameterDistribution()
print("Parameter distribution            =", parameterDistribution)



"""
Define a distribution from quantiles
====================================
"""

# %%
# In this example we are going to estimate a parametric distribution by numerical optimization of some quantiles.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We need as many quantile values as there are parameters
# of the distribution.
# For example, for a Normal distribution, the two parameters are the mean and the
# standard deviation.
# This implies that two quantiles are required to estimate the parameters
# of a Normal distribution.
# The values of the parameters :math:`\mu`, :math:`\sigma` will be used as the reference to assess the optimization.
dist_ref = ot.Normal(17.0, 34.0)
dist_ref.setDescription(["reference"])
p1 = 0.05
p2 = 0.95
q1 = dist_ref.computeQuantile(p1)[0]
q2 = dist_ref.computeQuantile(p2)[0]
print(q1, q2)

# %%
# Fit a Normal distribution from these quantiles by numerical optimization.
# Ensure we get the same values as the reference.
factory = ot.QuantileMatchingFactory(ot.Normal(), [p1, p2])
dist_estim = factory.buildFromQuantiles([q1, q2])
dist_estim.setDescription(["estimated"])
print(dist_estim)

# %%
# Graphically validate if the estimated distribution verifies the imposed quantiles.
graph = dist_estim.drawCDF()
curve_q1 = ot.Curve([q1, q1, 0.0], [0.0, p1, p1])
curve_q2 = ot.Curve([q2, q2, 0.0], [0.0, p2, p2])
graph.add(curve_q1)
graph.add(curve_q2)
text_q1 = ot.Text([[q1, -0.1]], [r"$q_1$"])
text_p1 = ot.Text([[0.0, p1]], [r"$p_1$"])
graph.add(text_q1)
graph.add(text_p1)
text_q2 = ot.Text([[q2, -0.1]], [r"$q_2$"])
text_p2 = ot.Text([[0.0, p2]], [r"$p_2$"])
graph.add(text_q2)
graph.add(text_p2)
_ = otv.View(graph)

# %%
# It is also possible to define a Histogram distribution from quantiles.
# As it is a non-parametric distribution we can define as many quantiles as needed.
N = 10
probabilities = [(i + 1) / N for i in range(N)]
probabilities[-1] = 1.0 - 1e-3
quantiles = [dist_ref.computeQuantile(pi)[0] for pi in probabilities]

# %%
# The service is part of the :class:`~openturns.HistogramFactory` class.
# We also need to define the lower bound of the Histogram to build the distribution.
lowerBound = quantiles[0] - 10.0
histo_quant = ot.HistogramFactory().buildFromQuantiles(
    lowerBound, probabilities, quantiles
)

# %%
# Graphically validate if the estimated distribution verifies the imposed quantiles.
# We can see that the CDF of the estimated Histogram matches all the quantile dots.
histo_quant.setDescription(["estimated"])
graph = histo_quant.drawCDF()
curve_qi = ot.Cloud(quantiles, probabilities)
curve_qi.setLegend("quantiles")
graph.add(curve_qi)
_ = otv.View(graph)

otv.View.ShowAll()



"""
Generate low discrepancy sequences
==================================
"""

# %%
# In this examples we are going to expose the available low discrepancy sequences in order to approximate some integrals.
#
# The following low-discrepancy sequences are available:
#
# - Sobol
# - Faure
# - Halton
# - reverse Halton
# - Haselgrove
#
# To illustrate these sequences we generate their first 1024 points and compare with the sequence obtained from
# the pseudo random generator (Merse Twister) as the latter has a higher discrepancy.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# 1. Sobol sequence
dimension = 2
size = 1024
sequence = ot.SobolSequence(dimension)
sample = sequence.generate(size)
graph = ot.Graph("", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# 2. Halton sequence
dimension = 2
sequence = ot.HaltonSequence(dimension)
sample = sequence.generate(size)
graph = ot.Graph("Halton", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# 3. Halton sequence in high dimension: bad filling in upper dimensions
dimension = 20
sequence = ot.HaltonSequence(dimension)
sample = sequence.generate(size).getMarginal([dimension - 2, dimension - 1])
graph = ot.Graph(
    "Halton (" + str(dimension - 2) + "," + str(dimension - 1) + ")",
    "dim " + str(dimension - 2),
    "dim " + str(dimension - 1),
    True,
    "",
)
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# 4. Scrambled Halton sequence in high dimension
dimension = 20
sequence = ot.HaltonSequence(dimension)
sequence.setScrambling("RANDOM")
sample = sequence.generate(size).getMarginal([dimension - 2, dimension - 1])
graph = ot.Graph(
    "Halton (" + str(dimension - 2) + "," + str(dimension - 1) + ")",
    "dim " + str(dimension - 2),
    "dim " + str(dimension - 1),
    True,
    "",
)
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# 5. Reverse Halton sequence
dimension = 2
sequence = ot.ReverseHaltonSequence(dimension)
sample = sequence.generate(size)
print(
    "discrepancy=",
    ot.LowDiscrepancySequenceImplementation.ComputeStarDiscrepancy(sample),
)
graph = ot.Graph("Reverse Halton", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# 6. Haselgrove sequence
dimension = 2
sequence = ot.HaselgroveSequence(dimension)
sample = sequence.generate(size)
graph = ot.Graph("Haselgrove", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# Compare with uniform random sequence
distribution = ot.JointDistribution([ot.Uniform(0.0, 1.0)] * 2)
sample = distribution.getSample(size)
print(
    "discrepancy=",
    ot.LowDiscrepancySequenceImplementation.ComputeStarDiscrepancy(sample),
)
graph = ot.Graph("Mersenne Twister", "", "")
cloud = ot.Cloud(sample)
graph.add(cloud)
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Perform stepwise regression
===========================
In this example we perform the selection of the most suitable function basis for a linear regression model with the help of the stepwise algorithm.

We consider te so-called Linthurst data set, which contains measures of aerial biomass (BIO) as well as 5 five physicochemical properties of the soil: salinity (SAL), pH, K, Na, and Zn.

The data set is taken from the book [rawlings2001]_
and is provided below:
"""

# %%
import openturns as ot
import openturns.viewer as otv
import numpy as np
import matplotlib.pyplot as plt
from openturns.usecases import linthurst

# %%
# We define the data.
#

# %%
ds = linthurst.Linthurst()
dimension = ds.data.getDimension() - 1

input_sample = ds.data[:, 1 : dimension + 1]
print("Input :")
print(input_sample[:5])
output_sample = ds.data[:, 0]
print("Output :")
print(output_sample[:5])

input_description = input_sample.getDescription()
output_description = output_sample.getDescription()

n = input_sample.getSize()
print("n = ", n)
dimension = ds.data.getDimension() - 1
print("dimension = ", dimension)


# %%
# Complete linear model
# ---------------------
#
# We consider a linear model with the purpose of predicting the aerial biomass as a function of the soil physicochemical properties,
# and we wish to identify the predictive variables which result in the most simple and precise linear regression model.
#
# We start by creating a linear model which takes into account all of the physicochemical variables present within the Linthrust data set.
#
# Let us consider the following linear model :math:`\tilde{Y} = a_0 + \sum_{i = 1}^{d} a_i X_i + \epsilon`. If all of the predictive variables
# are considered, the regression can be performed with the help of the :class:`~openturns.LinearModelAlgorithm` class.

# %%
algo_full = ot.LinearModelAlgorithm(input_sample, output_sample)
algo_full.run()
result_full = algo_full.getResult()
print("R-squared = ", result_full.getRSquared())
print("Adjusted R-squared = ", result_full.getAdjustedRSquared())

# %%
# Forward stepwise regression
# ---------------------------
#
# We now wish to perform the selection of the most important predictive variables through a stepwise algorithm.
#
# It is first necessary to define a suitable function basis for the regression. Each variable is associated to a univariate basis
# and an additional basis is used in order to represent the constant term :math:`a_0`.

# %%
functions = []
functions.append(ot.SymbolicFunction(input_description, ["1.0"]))
for i in range(dimension):
    functions.append(ot.SymbolicFunction(input_description, [input_description[i]]))
basis = ot.Basis(functions)
# %%
# Plese note that this example uses a linear basis with respect to the various predictors for the sake of clarity.
# However, this is not a necessity, and more complex and non linear relations between predictors may be considered
# (e.g., polynomial bases).


# %%
# We now perform a forward stepwise regression. We suppose having no information regarding the given data set, and therefore the set of minimal
# indices only contains the constant term (indexed by 0).
#
# The first regression is performed by relying on the Akaike Information Criterion (AIC), which translates into a penalty term equal to 2.
# In practice, the algorithm selects the functional basis subset that minimizes the AIC by iteratively adding the single function which provides
# the largest improvement until convergence is reached.

# %%
minimalIndices = [0]
direction = ot.LinearModelStepwiseAlgorithm.FORWARD
penalty = 2.0
algo_forward = ot.LinearModelStepwiseAlgorithm(
    input_sample, output_sample, basis, minimalIndices, direction
)
algo_forward.setPenalty(penalty)
algo_forward.run()
result_forward = algo_forward.getResult()
print("Selected basis: ", result_forward.getCoefficientsNames())
print("R-squared = ", result_forward.getRSquared())
print("Adjusted R-squared = ", result_forward.getAdjustedRSquared())

# %%
# With this first forward stepwise regression, the results show that the selected optimal basis contains a constant term,
# plus two linear terms depending respectively on the pH value (pH) and on the sodium concentration (Na).
#
# As can be expected, the :math:`R^2` value diminishes if compared to the regression on the entire basis, as the stepwise
# regression results in a lower number of predictive variables.  However, it can also be seen that the adjusted :math:`R^2`,
# which is a metric that also takes into account the ratio  between the amount of training data and the number of explanatory
# variables, is improved if compared to the complete model.

# %%
# Backward stepwise regression
# ----------------------------
#
# We now perform a backward stepwise regression, meaning that rather than iteratively adding predictive variables, we will be removing them,
# starting from the complete model.
# This regression is performed by relying on the Bayesian Information Criterion (BIC), which translates into a penalty term equal to :math:`log(n)`.

# %%
minimalIndices = [0]
direction = ot.LinearModelStepwiseAlgorithm.BACKWARD
penalty = np.log(n)
algo_backward = ot.LinearModelStepwiseAlgorithm(
    input_sample, output_sample, basis, minimalIndices, direction
)
algo_backward.setPenalty(penalty)
algo_backward.run()
result_backward = algo_backward.getResult()
print("Selected basis: ", result_backward.getCoefficientsNames())
print("R-squared = ", result_backward.getRSquared())
print("Adjusted R-squared = ", result_backward.getAdjustedRSquared())

# %%
# It is interesting to point out that although both approaches converge towards a model characterized by 2 predictive variables,
# the selected variables do not coincide.

# %%
# Both directions stepwise regression
# -----------------------------------
#
# A third available option consists in performing a stepwise regression in both directions, meaning that at each iteration the predictive variables
# can be either added or removed. In this case, a set of starting indices must be provided to the algorithm.

# %%
minimalIndices = [0]
startIndices = [0, 2, 3]
penalty = np.log(n)
direction = ot.LinearModelStepwiseAlgorithm.BOTH
algo_both = ot.LinearModelStepwiseAlgorithm(
    input_sample, output_sample, basis, minimalIndices, direction, startIndices
)
algo_both.setPenalty(penalty)
algo_both.run()
result_both = algo_both.getResult()
print("Selected basis: ", result_both.getCoefficientsNames())
print("R-squared = ", result_both.getRSquared())
print("Adjusted R-squared = ", result_both.getAdjustedRSquared())

# %%
# It is interesting to note that the basis varies depending on the selected set of starting indices, as is shown below.
# An informed initialization might therefore improve the model selection and the resulting regression

# %%
minimalIndices = [0]
startIndices = [0, 1]
penalty = np.log(n)
direction = ot.LinearModelStepwiseAlgorithm.BOTH
algo_both = ot.LinearModelStepwiseAlgorithm(
    input_sample, output_sample, basis, minimalIndices, direction, startIndices
)
algo_both.setPenalty(penalty)
algo_both.run()
result_both = algo_both.getResult()
print("Selected basis: ", result_both.getCoefficientsNames())
print("R-squared = ", result_both.getRSquared())
print("Adjusted R-squared = ", result_both.getAdjustedRSquared())

# %%
# Graphical analyses
# ------------------
#
# Finally, we can rely on the :class:`~openturns.LinearModelAnalysis` class in order to analyze
# the predictive differences between the obtained models.

# %%
analysis_full = ot.LinearModelAnalysis(result_full)
analysis_full.setName("Full model")
analysis_forward = ot.LinearModelAnalysis(result_forward)
analysis_forward.setName("Forward selection")
analysis_backward = ot.LinearModelAnalysis(result_backward)
analysis_backward.setName("Backward selection")
fig = plt.figure(figsize=(12, 8))
for k, analysis in enumerate([analysis_full, analysis_forward, analysis_backward]):
    graph = analysis.drawModelVsFitted()
    ax = fig.add_subplot(3, 1, k + 1)
    ax.set_title(analysis.getName(), fontdict={"fontsize": 16})
    graph.setXTitle("Exact values")
    ax.xaxis.label.set_size(12)
    ax.yaxis.label.set_size(14)
    graph.setTitle("")
    v = otv.View(graph, figure=fig, axes=[ax])
plt.tight_layout()


# %%
# For illustrative purposes, we show the Bayesian Information Criterion (BIC) and Akaike Information Criterion (AIC) values
# which are computed during the iterations of the forward step-wise regression. Please note that in order to do
# so, we set the penalty parameter to a negligible value (meaning that the basis selection only takes into account the model likelihood,
# and not the number of parameters characterizing the linear model).

# %%
minimalIndices = [0]
penalty = 1e-10
direction = ot.LinearModelStepwiseAlgorithm.FORWARD

BIC = []
AIC = []
for iterations in range(1, 6):
    algo_forward = ot.LinearModelStepwiseAlgorithm(
        input_sample, output_sample, basis, minimalIndices, direction
    )
    algo_forward.setPenalty(penalty)
    algo_forward.setMaximumIterationNumber(iterations)
    algo_forward.run()
    result_forward = algo_forward.getResult()

    RSS = np.sum(
        np.array(result_forward.getSampleResiduals()) ** 2
    )  # Residual sum of squares
    LL = n * np.log(RSS / n)  # Log-likelihood
    BIC.append(LL + iterations * np.log(n))  # Bayesian Information Criterion
    AIC.append(LL + iterations * 2)  # Akaike Information Criterion
    print("Selected basis: ", result_forward.getCoefficientsNames())


plt.figure()
plt.plot(np.arange(1, 6), BIC, label="BIC")
plt.plot(np.arange(1, 6), AIC, label="AIC")
plt.xticks(np.arange(1, 6))
plt.xlabel("Basis size", fontsize=14)
plt.ylabel("Information criterion", fontsize=14)
plt.legend(fontsize=14)
plt.tight_layout()

# %%
# The graphic above shows that the optimal linear model in terms of compromise between prediction likelihood and model complexity
# should take into account the influence of 2 regression variables as well as the constant term. This is consistent with the results previously
# obtained.

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
import openturns.viewer as otv

analytical = ot.SymbolicFunction(["x"], ["2*x-8"])
levelset = ot.LevelSet(analytical, ot.Less(), 0.0)
f = ot.IndicatorFunction(levelset)

graph = f.draw(0.0, 10.0)
graph.setTitle(r"$y=\mathbb{1}_{2x-8<0}$")
otv.View(graph, figure_kw={"figsize": (8, 4)}, add_legend=True)



"""
Gaussian Process Regression: metamodel of the Branin-Hoo function
=================================================================
"""

# %%
# As a popular use case in optimization we briefly present the construction of a metamodel
# of the Branin (also referred to as Branin-Hoo) function.
#

# %%
from numpy import sqrt
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import branin_function


# %%
# We load the Branin-Hoo model from the usecases module and use the model (stored in `objectiveFunction`)
bm = branin_function.BraninModel()
model = bm.objectiveFunction

# %%
# We shall represent this 2-d function with isolines. We set the number of isolines to a maximum of 10
# thanks to the following `ResourceMap` key :
ot.ResourceMap.SetAsUnsignedInteger("Contour-DefaultLevelsNumber", 10)
graphBasic = model.draw([0.0, 0.0], [1.0, 1.0], [100] * 2)
view = otv.View(graphBasic)

# %%
# We get the values of all isolines :
levels = graphBasic.getDrawables()[0].getLevels()

# # %%
# # We now build fancy isolines :

# # Build a range of colors
# ot.ResourceMap.SetAsUnsignedInteger("Drawable-DefaultPalettePhase", len(levels))
# palette = ot.Drawable.BuildDefaultPalette(len(levels))

graphFineTune = ot.Graph("The exact Branin model", r"$x_1$", r"$x_2$")
graphFineTune.setDrawables([graphBasic.getDrawable(0)])
# graphFineTune.setLegendPosition("")  # Remove the legend


# %%
# We also represent the three minima of the Branin function with orange diamonds :
sample1 = ot.Sample([bm.xexact1, bm.xexact2, bm.xexact3])
cloud1 = ot.Cloud(sample1, "orange", "diamond", "First Cloud")
graphFineTune.add(cloud1)
# Draw the graph with the palette assigned to the contour
view = otv.View(graphFineTune)

#
# The values of the exact model at these points are :
print(bm.objectiveFunction(sample1))

# %%
# The Branin function has a global minimum attained at three different points. We shall build a
# metamodel of this function that presents the same behaviour.


# %%
# Optimization of GP hyperparameters and conditioning to data
# -----------------------------------------------------------
#
# We use the :class:`~openturns.GaussianProcessFitter` class to perform the GP fitter analysis.
# We first generate a design of experiments with LHS and store the input training points in `xdata`
experiment = ot.LHSExperiment(
    ot.JointDistribution([ot.Uniform(0.0, 1.0), ot.Uniform(0.0, 1.0)]),
    28,
    False,
    True,
)
xdata = experiment.generate()

# %%
# We also get the output training values :
ydata = bm.objectiveFunction(xdata)


# %%
# This use case is defined in dimension 2 and we use a constant basis for the trend estimation :
dimension = bm.dim
basis = ot.ConstantBasisFactory(dimension).build()

# %%
# We choose a squared exponential covariance model in dimension 2 :
covarianceModel = ot.SquaredExponential([0.1] * dimension, [1.0])

# %%
# We have all the components to build a GP fitter  algorithm and run it :
fitter_algo = ot.GaussianProcessFitter(xdata, ydata, covarianceModel, basis)
fitter_algo.setOptimizeParameters(True)
fitter_algo.run()
fitter_result = fitter_algo.getResult()
gpr_algo = ot.GaussianProcessRegression(fitter_result)
gpr_algo.run()
gpr_result = gpr_algo.getResult()
print(gpr_result)

# %%
# Metamodel visualization
# -----------------------
#
# We get the metamodel (mean of the conditioned Gaussian Process) and the variance estimation.
# We draw the GP metamodel of the Branin function.
gprMetamodel = gpr_result.getMetaModel()
gccc = ot.GaussianProcessConditionalCovariance(gpr_result)


graphBasic = gprMetamodel.draw([0.0, 0.0], [1.0, 1.0], [100] * 2)
# Take the first drawable as the only contour with multiple levels
contours = graphBasic.getDrawable(0).getImplementation()
contours.setColorBarPosition("")  # Hide the color bar
contours.setDrawLabels(True)  # Draw the labels
levels = contours.getLevels()

# Build a range of colors
ot.ResourceMap.SetAsUnsignedInteger("Drawable-DefaultPalettePhase", len(levels))
palette = ot.Drawable.BuildDefaultPalette(len(levels))

graphFineTune = ot.Graph("Branin metamodel (mean)", r"$x_1$", r"$x_2$")
graphFineTune.setDrawables([contours])
graphFineTune.setLegendPosition("")

# %%
# We also represent the location of the minima of the Branin function :
sample1 = ot.Sample([bm.xexact1, bm.xexact2, bm.xexact3])
cloud1 = ot.Cloud(sample1, "orange", "diamond", "First Cloud")
graphFineTune.add(cloud1)
view = otv.View(graphFineTune)

# %%
# We evaluate the metamodel at the minima locations :
print(gprMetamodel(sample1))

# %%
# Graphically, both the metamodel and the exact function look the same. The metamodel also has three
# basins around the minima and the value of the metamodel at each minimum location is comparable to
# the exact value of -0.979476. We have roughly two correct digits for each isoline.


# %%
# Standard deviation
# ------------------
#
# We finally take a look at the standard deviation in the :math:`[0,1] \times [0,1]` domain. It may be
# seen as a measure of the error of the metamodel.

# %%
# We discretize the domain with 22 points (`N` inside points and 2 endpoints) :
N = 20
inputData = ot.Box([N, N]).generate()

# %%
# We compute the conditional variance of the model and take the square root to get the deviation :
condCov = gccc.getConditionalMarginalVariance(inputData, 0)
condCovSd = sqrt(condCov)

# %%
# As we have previously done we build contours with the following levels ans labels :
levels = [0.01, 0.025, 0.050, 0.075, 0.1, 0.125, 0.150, 0.175]
contour = ot.Contour(N + 2, N + 2, condCovSd)
contour.setLevels(levels)
graphFineTune = ot.Graph("Standard deviation", r"$x_1$", r"$x_2$")
graphFineTune.setDrawables([contour])
graphFineTune.setLegendPosition("")

# %%
# We superimpose the training sample :
cloud = ot.Cloud(xdata)
cloud.setPointStyle("fcircle")
cloud.setColor("red")
graphFineTune.add(cloud)
view = otv.View(graphFineTune)

# %%
# We observe that the standard deviation is small in the center of the domain where we have enough
# data to learn the model.
# We can print the value of the variance at the first 5 training points (they all behave similarly) :
print(gccc.getConditionalMarginalVariance(xdata, 0)[0:5])

# %%
# These values are nearly zero which is expected as the metamodel interpolates data. The value being
# known it is not random anymore and the variance ought to be zero.

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()


def cleanScalar(inScalar):
    if abs(inScalar) < 1.0e-10:
        inScalar = 0.0
    return inScalar


def cleanPoint(inPoint):
    dim = inPoint.getDimension()
    for i in range(dim):
        if abs(inPoint[i]) < 1.0e-10:
            inPoint[i] = 0.0
    return inPoint


ot.PlatformInfo.SetNumericalPrecision(5)
# Instantiate one distribution object
for dim in range(1, 5):
    meanPoint = [0.0] * dim
    sigma = [1.0 + i for i in range(dim)]
    R = ot.CorrelationMatrix(dim)
    for i in range(1, dim):
        R[i, i - 1] = 0.5

    distribution = ot.Normal(meanPoint, sigma, R)

    distribution.setName("A normal distribution")
    description = ["Marginal " + str(1 + i) for i in range(dim)]
    distribution.setDescription(description)

    print("Parameters collection=", repr(distribution.getParametersCollection()))
    print("Standard representative=", distribution.getStandardRepresentative())

    print("Distribution ", repr(distribution))
    print("Distribution ", distribution)
    print("Covariance ", repr(distribution.getCovariance()))

    # Is this distribution elliptical ?
    print("Elliptical = ", distribution.isElliptical())

    # Is this distribution continuous ?
    print("Continuous = ", distribution.isContinuous())

    # Test for realization of distribution
    oneRealization = distribution.getRealization()
    print("oneRealization=", repr(oneRealization))

    # Define a point
    point = ot.Point(distribution.getDimension(), 0.5)
    print("Point= ", repr(point))

    # Show PDF and CDF of point
    eps = 1e-5

    # derivative of PDF with respect to its arguments
    DDF = distribution.computeDDF(point)
    print("ddf     =", repr(cleanPoint(DDF)))

    # PDF value
    LPDF = distribution.computeLogPDF(point)
    print("log pdf=%.6f" % LPDF)
    PDF = distribution.computePDF(point)
    print("pdf     =%.6f" % PDF)

    CF = distribution.computeCharacteristicFunction(point)
    print("characteristic function=%.6f+%.6fi" % (CF.real, CF.imag))
    LCF = distribution.computeLogCharacteristicFunction(point)
    print("log characteristic function=%.6f+%.6fi" % (LCF.real, LCF.imag))
    CDF = distribution.computeCDF(point)
    print("cdf=%.6f" % CDF)
    CCDF = distribution.computeComplementaryCDF(point)
    print("ccdf=%.6f" % CCDF)
    PDFgr = distribution.computePDFGradient(point)
    print("pdf gradient     =", repr(cleanPoint(PDFgr)))

    # quantile
    if dim < 4:
        quantile = distribution.computeQuantile(0.95)
        print("quantile=", repr(quantile))
        print("cdf(quantile)=%.6f" % distribution.computeCDF(quantile))
        # Get 95% survival function
        inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
        print("InverseSurvival=", repr(inverseSurvival))
        print(
            "Survival(inverseSurvival)=%.6f"
            % distribution.computeSurvivalFunction(inverseSurvival)
        )
    print("entropy=%.6f" % distribution.computeEntropy())
    # Confidence regions
    if distribution.getDimension() <= 2:
        (
            interval,
            threshold,
        ) = distribution.computeMinimumVolumeIntervalWithMarginalProbability(0.95)
        print("Minimum volume interval=", interval)
        print("threshold=", ot.Point(1, threshold))
        levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
        print("Minimum volume level set=", levelSet)
        print("beta=", ot.Point(1, beta))
        (
            interval,
            beta,
        ) = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)
        print("Bilateral confidence interval=", interval)
        print("beta=", ot.Point(1, beta))
        (
            interval,
            beta,
        ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
            0.95, False
        )
        print("Unilateral confidence interval (lower tail)=", interval)
        print("beta=", ot.Point(1, beta))
        (
            interval,
            beta,
        ) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(
            0.95, True
        )
        print("Unilateral confidence interval (upper tail)=", interval)
        print("beta=", ot.Point(1, beta))

    mean = distribution.getMean()
    print("mean=", repr(mean))
    standardDeviation = distribution.getStandardDeviation()
    print("standard deviation=", repr(standardDeviation))
    skewness = distribution.getSkewness()
    print("skewness=", repr(skewness))
    kurtosis = distribution.getKurtosis()
    print("kurtosis=", repr(kurtosis))
    roughness = distribution.getRoughness()
    print("roughness=%.6f" % roughness)
    covariance = distribution.getCovariance()
    print("covariance=", repr(covariance))
    parameters = distribution.getParametersCollection()
    print("parameters=", repr(parameters))

    # Specific to this distribution
    beta = point.normSquare()
    densityGenerator = distribution.computeDensityGenerator(beta)
    print("density generator=%.6f" % densityGenerator)

    print(
        "pdf via density generator=%.6f"
        % ot.EllipticalDistribution.computePDF(distribution, point)
    )
    densityGeneratorDerivative = distribution.computeDensityGeneratorDerivative(beta)
    print("density generator derivative     =%.6f" % densityGeneratorDerivative)
    print(
        "density generator derivative (FD)=%.6f"
        % cleanScalar(
            (
                distribution.computeDensityGenerator(beta + eps)
                - distribution.computeDensityGenerator(beta - eps)
            )
            / (2.0 * eps)
        )
    )
    densityGeneratorSecondDerivative = (
        distribution.computeDensityGeneratorSecondDerivative(beta)
    )
    print(
        "density generator second derivative     =%.6f"
        % densityGeneratorSecondDerivative
    )
    print(
        "density generator second derivative (FD)=%.6f"
        % cleanScalar(
            (
                distribution.computeDensityGeneratorDerivative(beta + eps)
                - distribution.computeDensityGeneratorDerivative(beta - eps)
            )
            / (2.0 * eps)
        )
    )

    # Compute the radial CDF
    radius = 2.0
    print(
        "Radial CDF(%.6f" % radius,
        ")=%.6f" % distribution.computeRadialDistributionCDF(radius),
    )

    x = 0.6
    y = [0.2] * (dim - 1)
    print("conditional PDF=%.6f" % distribution.computeConditionalPDF(x, y))
    print("conditional CDF=%.6f" % distribution.computeConditionalCDF(x, y))
    print("conditional quantile=%.6f" % distribution.computeConditionalQuantile(x, y))
    pt = ot.Point([i + 1.5 for i in range(dim)])
    print(
        "sequential conditional PDF=",
        distribution.computeSequentialConditionalPDF(point),
    )
    resCDF = distribution.computeSequentialConditionalCDF(pt)
    print("sequential conditional CDF(", pt, ")=", resCDF)
    print(
        "sequential conditional quantile(",
        resCDF,
        ")=",
        distribution.computeSequentialConditionalQuantile(resCDF),
    )

    # Extract the marginals
    for i in range(dim):
        margin = distribution.getMarginal(i)
        print("margin=", repr(margin))
        print("margin PDF=%.6f" % margin.computePDF([0.5]))
        print("margin CDF=%.6f" % margin.computeCDF([0.5]))
        print("margin quantile=", repr(margin.computeQuantile(0.95)))
        print("margin realization=", repr(margin.getRealization()))
    if dim >= 2:
        # Extract a 2-D marginal
        indices = [1, 0]
        print("indices=", repr(indices))
        margins = distribution.getMarginal(indices)
        print("margins=", repr(margins))
        print("margins PDF=%.6f" % margins.computePDF([0.5] * 2))
        print("margins CDF=%.6f" % margins.computeCDF([0.5] * 2))
        quantile = margins.computeQuantile(0.95)
        print("margins quantile=", repr(quantile))
        print("margins CDF(qantile)=%.6f" % margins.computeCDF(quantile))
        print("margins realization=", repr(margins.getRealization()))

    chol = distribution.getCholesky()
    invChol = distribution.getInverseCholesky()
    print("chol=", repr(chol))
    print("chol*t(chol)=", repr((chol * chol.transpose())))
    ott.assert_almost_equal((chol * invChol), ot.IdentityMatrix(dim))

    ot.Log.Show(ot.Log.TRACE)
    validation = ott.DistributionValidation(distribution)
    if dim > 2:
        validation.skipMoments()  # slow
        validation.skipCorrelation()  # slow
    validation.run()

# non-spd cov
dist = ot.Normal(
    [0] * 3, ot.CovarianceMatrix([[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]])
)
sample = dist.getSample(10)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.RandomGenerator.SetSeed(0)

distribution = ot.Triangular(1.0, 2.5, 4.0)
size = 10000
sample = distribution.getSample(size)
factory = ot.TriangularFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedTriangular = factory.buildAsTriangular(sample)
print("Triangular          =", distribution)
print("Estimated triangular=", estimatedTriangular)
estimatedTriangular = factory.buildAsTriangular()
print("Default triangular=", estimatedTriangular)
estimatedTriangular = factory.buildAsTriangular(distribution.getParameter())
print("Triangular from parameters=", estimatedTriangular)



"""
Validate a polynomial chaos
===========================
"""

# %%
#
# In this example, we show how to perform the validation plot of a polynomial chaos for the :ref:`Ishigami function <use-case-ishigami>`.

# %%
from openturns.usecases import ishigami_function
import openturns as ot
import openturns.viewer as otv

# %%
# Model description
# -----------------

# %%
# We load the Ishigami test function from the usecases module :
im = ishigami_function.IshigamiModel()

# %%
# The model contains the input distribution :math:`X=(X_1, X_2, X_3)` in
# `im.distribution` and the Ishigami function in `im.model`.
# We also have access to the input variable names with
input_names = im.distribution.getDescription()

# %%
N = 100
inputTrain = im.distribution.getSample(N)
outputTrain = im.model(inputTrain)

# %%
# Create the chaos
# ----------------

# %%
# We could use only the input and output training samples: in this case, the distribution of the input sample is computed by selecting the best distribution that fits the data.

# %%
chaosalgo = ot.FunctionalChaosAlgorithm(inputTrain, outputTrain)

# %%
# Since the input distribution is known in our particular case,
# we instead create the multivariate basis from the distribution,
# that is three independent variables :math:`X_1` , :math:`X_2` and :math:`X_3` .

# %%
multivariateBasis = ot.OrthogonalProductPolynomialFactory([im.X1, im.X2, im.X3])
totalDegree = 8
enumfunc = multivariateBasis.getEnumerateFunction()
basisSize = enumfunc.getBasisSizeFromTotalDegree(totalDegree)
adaptiveStrategy = ot.FixedStrategy(multivariateBasis, basisSize)

# %%
selectionAlgorithm = ot.LeastSquaresMetaModelSelectionFactory()
projectionStrategy = ot.LeastSquaresStrategy(
    inputTrain, outputTrain, selectionAlgorithm
)

# %%
chaosalgo = ot.FunctionalChaosAlgorithm(
    inputTrain, outputTrain, im.distribution, adaptiveStrategy, projectionStrategy
)

# %%
chaosalgo.run()
result = chaosalgo.getResult()
metamodel = result.getMetaModel()

# %%
# Validation of the metamodel
# ---------------------------

# %%
# In order to validate the metamodel, we generate a test sample.

# %%
n_valid = 1000
inputTest = im.distribution.getSample(n_valid)
outputTest = im.model(inputTest)
metamodelPredictions = metamodel(inputTest)
val = ot.MetaModelValidation(outputTest, metamodelPredictions)
r2Score = val.computeR2Score()[0]
r2Score

# %%
# The :math:`R^2` is very close to 1: the metamodel seems very accurate.

# %%
graph = val.drawValidation()
graph.setTitle("R2=%.2f%%" % (r2Score * 100))
view = otv.View(graph)

# %%
# The metamodel has a good predictivity, since the points are almost on the first diagonal.

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import math
from openturns.usecases import ishigami_function

ot.TESTPREAMBLE()

# Test integrate
im = ishigami_function.IshigamiModel()
sampleSize = 100000
experiment = ot.MonteCarloExperiment(im.distribution, sampleSize)
integration = ot.ExperimentIntegration(experiment)
approximatedOutputMean = integration.integrate(im.model)
rtol = 0.0
atol = 1.0 / math.sqrt(sampleSize)
assert approximatedOutputMean.getDimension() == 1
ott.assert_almost_equal(approximatedOutputMean[0], im.expectation, rtol, atol)

# Test computeL2Norm
im = ishigami_function.IshigamiModel()
centeredIshigamiFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3"], ["sin(x1) + 7 * (sin(x2)) ^ 2 + 0.1 * x3^4 * sin(x1) - 3.5"]
)
functionNorm = integration.computeL2Norm(centeredIshigamiFunction)
exactFunctionNorm = math.sqrt(im.variance)
rtol = 0.0
atol = 10.0 / math.sqrt(sampleSize)
ott.assert_almost_equal(functionNorm[0], exactFunctionNorm, rtol, atol)

# Test compute error
im = ishigami_function.IshigamiModel()
partOfIshigamiFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3"], ["7 * (sin(x2)) ^ 2 + 0.1 * x3^4 * sin(x1)"]
)
functionError = integration.computeL2Norm(im.model - partOfIshigamiFunction)
exactError = math.sqrt(0.5)
rtol = 0.0
atol = 10.0 / math.sqrt(sampleSize)
ott.assert_almost_equal(functionError[0], exactError, rtol, atol)

# Test integrate with multi-variate output
functionCollection = [im.model, im.model, im.model]
multivariateIshigami = ot.AggregatedFunction(functionCollection)
experiment = ot.MonteCarloExperiment(im.distribution, sampleSize)
integration = ot.ExperimentIntegration(experiment)
approximatedOutputMean = integration.integrate(multivariateIshigami)
rtol = 0.0
atol = 10.0 / math.sqrt(sampleSize)
assert approximatedOutputMean.getDimension() == 3
exactIntegral = [im.expectation] * 3
ott.assert_almost_equal(approximatedOutputMean, exactIntegral, rtol, atol)

# Test integrate with range
function = ot.SymbolicFunction(["x", "y", "z"], ["x^2+y^2+z^2"])
experiment = ot.GaussProductExperiment(
    ot.JointDistribution([ot.Uniform()] * 3), [4] * 3
)
integration = ot.ExperimentIntegration(experiment)
approximatedOutputMean = integration.integrate(
    function, ot.Interval([0.5] * 3, [1.0] * 3)
)
rtol = 1.0e-14
atol = 1.0e-14
ott.assert_almost_equal(approximatedOutputMean[0], 7.0 / 4.0, rtol, atol)



"""
Create a sparse chaos by integration
====================================
"""

# %%
#
# The goal of this example is to show how to create a sparse polynomial chaos
# expansion (PCE) when we estimate its coefficients by integration. We show
# how to use the :class:`~openturns.CleaningStrategy` class.
#
# Polynomial chaos expansion
# --------------------------
#
# Let :math:`g : \mathcal{D} \rightarrow \mathbb{R}` be a function
# where :math:`\mathcal{D} \subseteq \mathbb{R}^p` is the domain of :math:`g`.
# We consider :math:`\vect{X}` a random vector which
# probability density function is denoted by :math:`f`.
# We assume that :math:`g(\vect{X})` has a finite second order moment.
# Let :math:`T` be an iso-probabilistic transformation such that :math:`\vect{Z} = T(\vect{X})`
# follows a distribution uniquely defined by all its moments.
# Let :math:`h` be the function defined by:
#
# .. math::
#
#   h = g \circ T^{-1}.
#
# The polynomial chaos decomposition of :math:`h` with respect to the measure of
# :math:`\vect{Z}` is (see [blatman2009]_ page 73) :
#
# .. math::
#
#   h(\vect{z}) = \sum_{\vect{\alpha} \in \mathbb{N}^p}
#   a_{\vect{\alpha}} \psi_{\vect{\alpha}}(\vect{z})
#
# where :math:`\vect{\alpha} = (\alpha_1, ..., \alpha_p) \in \mathbb{N}^p`
# is a multi-index, :math:`a_{\vect{\alpha}} \in \mathbb{R}` is the
# coefficient,  :math:`\psi_{\vect{\alpha}} : \mathcal{\bar{X}} \rightarrow \mathbb{R}`
# is a multivariate polynomial.

# %%
#
# Truncated expansion
# -------------------
# In practice, we cannot consider an infinite series and must truncate
# the decomposition at a given order. Only a selection of coefficients must
# be kept. This leads to a subset of all possible multi-indices. In the
# remainder of this text, we call this selection the *multi-index set*.
#
# Several multi-index sets can be considered. A simple method is to truncate
# the polynomial up to a given maximum total degree :math:`d \in \mathbb{N}`.
# Let :math:`\mathcal{A}^{d}` be the multi-index set defined by
#
# .. math::
#
#   \mathcal{A}^{d} = \left\{ \vect{\alpha} \in \mathbb{N}^p
#   \; | \; \|\vect{\alpha}\|_1 \leq d\right\}
#
# where
#
# .. math::
#
#   \|\vect{\alpha}\|_d = \alpha_1 + ... + \alpha_p
#
# is the 1-norm of the multi-index :math:`\vect{\alpha}`.
# Therefore, the truncated polynomial chaos expansion is:
#
# .. math::
#
#   \widetilde{h}(\vect{z}) = \sum_{\vect{\alpha} \in \mathcal{A}^{d}}
#   a_{\vect{\alpha}} \psi_{\vect{\alpha}}(\vect{z}).
#
# In order to ensure a low error, we may choose a large value of the
# parameter :math:`P`. This, however, leads to a large number of
# coefficients :math:`\vect{\alpha} \in \mathcal{A}^{d}` to
# estimate. More precisely, the number of coefficients to estimate
# is (see [blatman2009]_ page 73) :
#
# .. math::
#
#   \textrm{card}\left(\mathcal{A}^{d}\right) = {p + d \choose d}
#   = \frac{(p + d)!}{p! d!}
#
# where :math:`p!` is the factorial number of :math:`p`.

# %%
#
# Low-rank polynomial chaos expansion
# -----------------------------------
# For any :math:`\vect{\alpha} \in \mathbb{N}^p`, let
# :math:`\|\vect{\alpha}\|_0` be the rank of the multi-index, that is,
# the number of non-zero components:
#
# .. math::
#
#   \|\vect{\alpha}\|_0 = \sum_{i = 1}^p \vect{1}_{\alpha_i > 0}
#
# where :math:`\vect{1}` is the indicator function.
# The multi-index set of maximum total degree :math:`d \in \mathbb{N}`
# and maximum rank :math:`j \in \mathbb{N}` is ([blatman2009]_ page 74):
#
# .. math::
#
#   \mathcal{A}^{d,j} = \left\{ \vect{\alpha} \in \mathbb{N}^p
#   \; | \; \|\vect{\alpha}\|_1 \leq d, \;
#   \; \|\vect{\alpha}\|_0 \leq j\right\}.
#
# Therefore, the rank-`j` polynomial chaos expansion is:
#
# .. math::
#
#   \widetilde{h}(\vect{z}) = \sum_{\vect{\alpha} \in
#   \mathcal{A}^{d,j}} a_{\vect{\alpha}}
#   \psi_{\vect{\alpha}}(\vect{z}).
#
# The rank is now a hyperparameter of the model: [blatman2009]_ suggests
# to use :math:`j = 2, 3, 4`. An example of low-rank PCE for the G-Sobol'
# function is given in [blatman2009]_ page 75.
#
# *Note.* It is currently not possible to create a low-rank PCE.

# %%
#
# Model selection
# ---------------
# If :math:`\textrm{Card}\left(\mathcal{A}^{d}\right)` is large, many coefficients
# may be poorly estimated, which may reduce the quality of the metamodel. We may
# want to select a subset of the coefficients which best predict the output.
# In other words, we may compute a subset:
#
# .. math::
#
#   \mathcal{A} \subseteq \mathcal{A}^{d}
#
# such that ([blatman2009]_ page 86) :
#
# .. math::
#
#   \widetilde{h}(\vect{z}) = \sum_{\vect{\alpha} \in \mathcal{A}}
#   a_{\vect{\alpha}} \psi_{\vect{\alpha}}(\vect{z})
#
# An enumeration rule is a function from the set of integers :math:`k` to
# the corresponding set of multi-indices :math:`\vect{\alpha}`. More
# precisely, let :math:`r : \mathbb{N} \rightarrow \mathbb{N}^p` be the
# function such that :
#
# .. math::
#
#   r(k) = \vect{\alpha}
#
# for any :math:`k \geq 0`.
# Let :math:`K \in \mathbb{N}` be a parameter representing the number of
# coefficients considered in the selection. Given an enumeration rule for
# the multi-indices :math:`\vect{\alpha}`, at most :math:`K` multi-indices
# will be considered. Let :math:`\mathcal{A}_K` be the corresponding multi-index set :
#
# .. math::
#
#   \mathcal{A}_K = \left\{ \vect{\alpha}
#   \; | \; r^{-1}(\vect{\alpha}) = k \leq K \right\}.
#
#
# Let :math:`\epsilon > 0` be a parameter representing the minimum relative
# value of a significant coefficient :math:`a_{\vect{\alpha}}`.
# The :class:`~openturns.CleaningStrategy` uses the following criterion to select the coefficients :
#
# .. math::
#
#   \mathcal{A}_\epsilon =
#   \left\{
#   |a_{\vect{\alpha}}| \geq \epsilon \max_{ a_{\vect{\alpha}}
#   \in \mathcal{A}_K } |a_{\vect{\alpha}}| \right\}
#
# where :math:`\epsilon` is the significance factor, which by default is
# :math:`\epsilon = 10^{-4}`. This rule selects only the coefficients which
# are significantly different from zero.

# %%
#
# Sparsity index
# --------------
# The sparsity index of a multi-index set is the ratio of the cardinality of
# the multi-index set to the cardinality of the multi-index set of the
# equivalent multi-index with maximum total degree. For a given multi-index
# set :math:`\mathcal{A}`, let :math:`d` be the maximum 1-norm of multi-indices
# in the set :
#
# .. math::
#
#   d := \textrm{max}_{\vect{\alpha} \in \mathcal{A}}
#   \|\vect{\alpha}\|_1.
#
# The index of sparsity of :math:`\mathcal{A}` is ([blatman2009]_  eq. 4.42 page 86) :
#
# .. math::
#
#   \textrm{IS}(\mathcal{A})
#   = \frac{\textrm{card}(\mathcal{A})}{\textrm{card}\left(\mathcal{A}^d\right)}.
#
#
# *Note.* The index of sparsity as defined by [blatman2009]_ is close to zero when
# the model is very sparse. The following complementary indicator is close
# to 1 when the model is very sparse:
#
# .. math::
#
#   \textrm{IS}_{\textrm{c}}(\mathcal{A})
#   = 1 - \frac{\textrm{card}(\mathcal{A})}{\textrm{card}\left(\mathcal{A}^d\right)}.
#

# %%
import openturns as ot
import openturns.viewer as otv
from openturns.usecases import ishigami_function
import itertools


# %%
#
# The following function takes a polynomial chaos result as input and prints a
# given maximum number of coefficients of this polynomial. It can take into account
# a threshold, so that we can avoid to print coefficients which are very close to zero.


# %%
def printCoefficientsTable(
    polynomialChaosResult, maximum_number_of_printed_coefficients=10, threshold=0.0
):
    """
    Print the coefficients of the polynomial chaos.

    Parameters
    ----------
    polynomialChaosResult : ot.FunctionalChaosResult
        The polynomial chaos expansion.
    maximum_number_of_printed_coefficients : int
        The maximum number of printed coefficients.
    threshold : float, strictly positive
        If a coefficient has an absolute value strictly greater than the
        threshold, it is printed.
    """
    basis = polynomialChaosResult.getOrthogonalBasis()
    coefficients = polynomialChaosResult.getCoefficients()
    enumerate_function = basis.getEnumerateFunction()
    indices = polynomialChaosResult.getIndices()
    nbcoeffs = indices.getSize()
    print("Total number of coefficients : ", nbcoeffs)
    print("# Indice, Multi-indice, Degree : Value")
    print_index = 0
    for k in range(nbcoeffs):
        multiindex = enumerate_function(indices[k])
        degree = sum(multiindex)
        c = coefficients[k][0]
        if abs(c) > threshold:
            print("#%d, %s (%s) : %s" % (k, multiindex, degree, c))
            print_index += 1
        if print_index > maximum_number_of_printed_coefficients:
            break
    return


# %%
#
# The next function computes the polynomial chaos :math:`R^2` score using simple validation
# on a test sample generated by Monte-Carlo sampling. The actual computation
# is performed by the :class:`~openturns.MetaModelValidation` class.


# %%
def compute_polynomial_chaos_R2(
    polynomialchaos_result, g_function, input_distribution, n_valid=1000
):
    """
    Compute the R2 score of the polynomial chaos.


    Parameters
    ----------
    polynomialChaosResult : ot.FunctionalChaosResult
        The polynomial chaos expansion.
    g_function : ot.Function
        The function.
    input_distribution : ot.Distribution
        The input distribution.
    n_valid : int
        The number of simulations to compute the R2 score.

    Returns
    -------
    r2Score : float
        The R2 score
    """
    ot.RandomGenerator.SetSeed(1976)
    metamodel = polynomialchaos_result.getMetaModel()
    inputTest = input_distribution.getSample(n_valid)
    outputTest = g_function(inputTest)
    metamodelPredictions = metamodel(inputTest)
    val = ot.MetaModelValidation(outputTest, metamodelPredictions)
    r2Score = val.computeR2Score()[0]
    return r2Score


# %%
#
# The following function creates a validation plot using the
# :meth:`~openturns.MetaModelValidation.drawValidation` method
# of the :class:`~openturns.MetaModelValidation` class.


# %%
def draw_polynomial_chaos_validation(
    polynomialchaos_result, g_function, input_distribution, n_valid=1000
):
    """
    Validate the polynomial chaos.

    Create the validation plot.


    Parameters
    ----------
    polynomialChaosResult : ot.FunctionalChaosResult
        The polynomial chaos expansion.
    g_function : ot.Function
        The function.
    input_distribution : ot.Distribution
        The input distribution.
    n_valid : int
        The number of simulations to compute the R2 score.

    Returns
    -------
    view : ot.View
        The plot.
    """
    metamodel = polynomialchaos_result.getMetaModel()
    inputTest = input_distribution.getSample(n_valid)
    outputTest = g_function(inputTest)
    metamodelPredictions = metamodel(inputTest)
    val = ot.MetaModelValidation(outputTest, metamodelPredictions)
    r2Score = val.computeR2Score()[0]
    graph = val.drawValidation()
    graph.setTitle("R2=%.2f%%" % (r2Score * 100))
    view = otv.View(graph, figure_kw={"figsize": (5.0, 4.0)})
    return view


# %%
#
# We consider the Ishigami model.
# Its three inputs are i.i.d. random variables that follow the uniform distribution on the
# :math:`[-\pi, \pi]` interval. This is an interesting example for our
# purpose because it is highly non-linear, so that a high polynomial degree
# will be required in order to produce a polynomial chaos expansion with :math:`R^2`
# score sufficiently close to 1.

# %%
im = ishigami_function.IshigamiModel()

im.distribution.setDescription(["X_0", "X_1", "X_2"])
im.model.setOutputDescription(["Y"])
print(im.distribution)

# %%
#
# Then we create the multivariate basis onto which the function is expanded.
# By default, it is associated with the linear enumeration rule. Since our
# marginals are uniform, the :class:`~openturns.OrthogonalProductPolynomialFactory` class
# produces Legendre polynomials.
# In order to create the multivariate basis of polynomials, we must specify the
# number of functions in the basis. In this particular case, we compute that
# number depending on the total degree.
# The :meth:`~openturns.EnumerateFunction.getMaximumDegreeStrataIndex` method
# of the enumeration function computes the number of layers necessary to achieve
# that total degree. Then the number of functions up to that layer is computed
# with the :meth:`~openturns.EnumerateFunction.getStrataCumulatedCardinal` method.

# %%
dimension = im.distribution.getDimension()
multivariateBasis = ot.OrthogonalProductPolynomialFactory(
    [im.distribution.getMarginal(i) for i in range(dimension)]
)

totalDegree = 5  # Polynomial degree
enumerate_function = multivariateBasis.getEnumerateFunction()
strataIndex = enumerate_function.getMaximumDegreeStrataIndex(totalDegree)
print("strataIndex = ", strataIndex)
number_of_terms_in_basis = enumerate_function.getStrataCumulatedCardinal(strataIndex)
print("number_of_terms_in_basis = ", number_of_terms_in_basis)
adaptiveStrategy = ot.FixedStrategy(multivariateBasis, number_of_terms_in_basis)
print(adaptiveStrategy)

# %%
#
# We compute the coefficients using a multivariate tensor product Gaussian
# quadrature rule. Since the coefficients are computed in the standardized
# space, we first use the `getMeasure` method of the multivariate basis in
# order to get that standardized distribution. Then we use the
# :class:`~openturns.GaussProductExperiment` class to create the quadrature,
# using 6 nodes on each
# of the dimensions.

# %%
standard_distribution = multivariateBasis.getMeasure()
print(standard_distribution)

marginal_number_of_nodes = 6
dim_input = im.model.getInputDimension()
marginalDegrees = [marginal_number_of_nodes] * dim_input
experiment = ot.GaussProductExperiment(im.distribution, marginalDegrees)
X, W = experiment.generateWithWeights()
Y = im.model(X)
print("Sample size = ", X.getSize())

# %%
#
# We see that 216 nodes are involved in this quadrature rule, which is the result
# of :math:`6^3 = 216`.
#
# In the next cell, we compute the coefficients of the polynomial chaos expansion
# using integration.

# %%
projectionStrategy = ot.IntegrationStrategy()
chaosalgo = ot.FunctionalChaosAlgorithm(
    X, W, Y, im.distribution, adaptiveStrategy, projectionStrategy
)
chaosalgo.run()
result = chaosalgo.getResult()

# %%
#
# We now validate the metamodel by drawing the validation graph. We see that
# many points are close to the red test line, which indicates that the
# predictions of the polynomial chaos expansion are close to the output
# observations from the model.

# %%
view = draw_polynomial_chaos_validation(result, im.model, im.distribution)

# %%
#
# In order to have a closer look on the result, we use the
# `printCoefficientsTable` function in order to print the first 10 coefficients.

# %%
printCoefficientsTable(result)


# %%
#
# We see that there are 56 coefficients in the metamodel and that many of
# these coefficients are close to zero.
#
# If we print only coefficients greater than :math:`10^{-14}`, we see that
# only a fraction of them are significant and that these significant coefficients
# have a relatively large polynomial degree.

# %%
printCoefficientsTable(result, threshold=1.0e-14)


# %%
#
# The previous experiments suggest to keep only the coefficients which are
# significant in the model: this is the topic of the next section.

# %%
#
# Use a model selection method
# ----------------------------


# %%
#
# The `CleaningStrategy` has the following algorithm. On input, it considers
# only the first `maximumConsideredTerms` coefficients
# :math:`a_{\vect{\alpha}}`. On output it selects the `mostSignificant`
# most significant coefficients. To do this, it uses the
# `significanceFactor` parameter.
#
# The following function will help to create a sparse PCE using the
# `CleaningStrategy`. It takes into account the number of considered coefficients
# in the expansion, the number of significant coefficients to keep and the
# relative factor and returns the :math:`R^2` score.


# %%
def compute_cleaning_PCE(
    maximumConsideredTerms, mostSignificant, significanceFactor, verbose=False
):
    """
    Compute a PCE using CleaningStrategy.

    Parameters
    ----------
    maximumConsideredTerms : int
        The maximum number of coefficients considered by the algorithm during
        intermediate steps.

    mostSignificant : int
        The maximum number of coefficients selected by the algorithm in
        the final PCE.

    significanceFactor : float
        The relative part of any coefficient with respect to the maximum
        coefficient.

    verbose : bool
        If set to True, print intermediate messages.

    Returns
    -------
    R2 : float
        The R2 score
    """
    adaptiveStrategy = ot.CleaningStrategy(
        multivariateBasis,
        maximumConsideredTerms,
        mostSignificant,
        significanceFactor,
    )
    chaosalgo = ot.FunctionalChaosAlgorithm(
        X, W, Y, im.distribution, adaptiveStrategy, projectionStrategy
    )
    chaosalgo.run()
    result = chaosalgo.getResult()
    score_R2 = compute_polynomial_chaos_R2(result, im.model, im.distribution)
    if verbose:
        print("R2 = %.2f%%" % (100.0 * score_R2))
        printCoefficientsTable(result)
    return score_R2


# %%
#
# In the next cell, we consider at most 500 coefficients and keep only
# the 5 most significant coefficients. The factor is set to a relatively low value.

# %%
maximumConsideredTerms = 500
mostSignificant = 5
significanceFactor = 1.0e-10
score_R2 = compute_cleaning_PCE(
    maximumConsideredTerms, mostSignificant, significanceFactor, verbose=True
)


# %%
#
# We see that when we keep only 5 coefficients among the first 500 ones,
# these coefficients have a very high polynomial degree. Indeed, it occurs that
# these poorly estimated coefficients have a high absolute value. Hence, the
# criteria selects them as significant coefficients, which leads to a poor metamodel.
#
# Let us reduce the number of considered coefficients and increase the number
# of selected coefficients.

# %%
maximumConsideredTerms = 56
mostSignificant = 10
significanceFactor = 1.0e-10
score_R2 = compute_cleaning_PCE(
    maximumConsideredTerms, mostSignificant, significanceFactor, verbose=True
)


# %%
#
# When we keep only 10 coefficients among the first 56 ones, the polynomial
# chaos metamodel is much better: the coefficients are associated with a low
# polynomial degree, so that the quadrature rule estimates them with greater accuracy.
#
# We would like to know which combination is the best. In the following loop, we
# consider the maximum number of considered coefficients from 1 to 500 and the
# number of selected coefficients from 1 to 30. In order to produce the
# combinations, we use the `product` function from the `itertools` module.
# For each combination, we compute the :math:`R^2` score and select the
# combination with the highest :math:`R^2` coefficient. As shown in [muller2016]_
# page 268, the computed :math:`R^2` may be optimistic, but this is not the
# point of the current example.

# %%
#
maximumConsideredTerms_list = list(range(1, 500, 50))
mostSignificant_list = list(range(1, 30, 5))
iterator = itertools.product(maximumConsideredTerms_list, mostSignificant_list)
best_R2_score = 0.0
best_parameters = []
for it in iterator:
    maximumConsideredTerms, mostSignificant = it
    score_R2 = compute_cleaning_PCE(
        maximumConsideredTerms, mostSignificant, significanceFactor
    )
    if score_R2 > best_R2_score:
        best_R2_score = score_R2
        best_parameters = [maximumConsideredTerms, mostSignificant]

print("Best R2 = %.2f%%" % (100.0 * best_R2_score))

maximumConsideredTerms, mostSignificant = best_parameters
print("Number of considered coefficients : ", maximumConsideredTerms)
print("Number of selected coefficients : ", mostSignificant)


# %%
#
# We see that the best solution could be to select at most 16 significant
# coefficients among the first 101 ones. Let us see the :math:`R^2` score and the
# coefficients in this situation.

# %%
score_R2 = compute_cleaning_PCE(
    maximumConsideredTerms, mostSignificant, significanceFactor, verbose=True
)


# %%
#
# These parameters lead to a total number of coefficients equal to 12. Among
# the 16 most significant coefficients, only 12 satisfy the criteria. Most of
# the coefficients have a small polynomial degree although some have a total
# degree as large as 7.

# %%
#
# Intermediate steps of the algorithm
# -----------------------------------
#
# If we set the `verbose` optional input argument of the
# `compute_cleaning_PCE` function to `True`, then intermediate messages are
# printed in the Terminal. For each step of
# the adaptivity algorithm, the code prints some of the internal parameters
# of the algorithm. The data structure uses several variables that we now describe.
#
# - `Psi_k_p_` : the collection of functions in the current active polynomial multi-index set,
# - `I_p_` : the list of indices of the selected coefficients based according to the enumeration rule,
# - `addedPsi_k_ranks_` : the list of indices to add the multi-index set,
# - `removedPsi_k_ranks_` : the list of indices to remove to the multi-index set,
# - `conservedPsi_k_ranks_` : the index of the first polynomial in the selected multi-index set,
# - `currentVectorIndex_` : the current value of the index in the full multi-index set, according to the enumeration rule.
#
# Each time the selection method is called, it is passed a
# coefficient :math:`a_{\vect{\alpha}}` which is a new candidate to be
# considered by the algorithm. The first time the method is evaluated, the
# active multi-index set is empty, so that it must be filled with the first
# coefficients in the multi-index set, according to the enumeration rule. The
# second time (and up to the end of the algorithm), the candidate coefficient
# is considered to be added to the multi-index set.
#
# Executing the function prints messages that we can process to produce the
# following listing. On each step, we print the list of integers corresponding
# to the indices of the coefficients in the active multi-index set.

# %%
#
# .. code-block:: python
#
#    Step 1:  [0, 1, 7, 10, 15, 16]
#    Step 2:  [0, 1, 7, 10, 15, 17]
#    Step 3:  [0, 1, 7, 10, 15, 18]
#    Step 4:  [0, 1, 7, 10, 15, 19]
#    Step 5:  [0, 1, 7, 10, 15, 20]
#    Step 6:  [0, 1, 7, 10, 15, 21]
#    Step 7:  [0, 1, 7, 10, 15, 22]
#    Step 8:  [0, 1, 7, 10, 15, 23]
#    Step 9:  [0, 1, 7, 10, 15, 24]
#    Step 10: [0, 1, 7, 10, 15, 25]
#    [...]
#    Step 15: [0, 1, 7, 10, 15, 30]
#    Step 16: [0, 1, 7, 10, 15, 30, 31]
#    [...]
#    Step 20: [0, 1, 7, 10, 15, 30, 35]
#    Step 21: [0, 1, 7, 10, 15, 30, 35, 36]
#    [...]
#    Step 25: [0, 1, 7, 10, 15, 30, 35, 40]
#    Step 26: [0, 1, 7, 10, 15, 30, 35, 40, 41]
#    [...]
#    Step 34: [0, 1, 7, 10, 15, 30, 35, 40, 49]
#    Step 35: [0, 1, 7, 10, 15, 30, 35, 40, 49, 50]
#    [...]
#    Step 69: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84]
#    Step 70: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 85]
#    [...]
#    Step 74: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89]
#    Step 75: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 90]
#    [...]
#    Step 83: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98]
#    Step 84: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98, 99]
#    Step 85: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98, 100]
#    Step 86: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98]
#    Step 87: [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98]

# %%
#
# The previous text (and a detailed analysis of the output file) allows one to
# understand what exactly happens in the algorithm. To understand each step,
# note that the significant threshold is equal to :math:`\epsilon = 10^{-10}`.
#
# - During the initialization, the initial basis is empty and filled with the
#   indices [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15].
#   The coefficient with index equal to 0 corrresponds to the mean of the
#   polynomial chaos expansion. As a result, we always want to select that
#   coefficient in the active set.
#   Hence, the calculation of the largest coefficient in absolute value is
#   performed from the indices from 1 to 15.
#   The largest coefficient in absolute value is :math:`a_1 = 1.625`, which leads
#   to the threshold :math:`\epsilon |a_1| = 1.625 \times 10^{-10}`. Most of
#   the considered coefficient are, however, too close to zero. This is why only
#   the coefficients [0, 1, 7, 10, 15] are kept in the basis. The
#   corresponding coefficients are [3.505, 1.625,-0.6414, -1.291, 1.372].
# - On step 1, the candidate index 16 is considered. Its coefficient is
#   :math:`a_{16} = -1.197 \times 10^{-15}`, which is much too low to be
#   selected. Hence, the basis is unchanged and the active multi-index set
#   is [0, 1, 7, 10, 15] on the end of this step.
# - From the step 3 to the step 15, the active multi-index set is unchanged,
#   because no considered coefficient becomes greater than the threshold.
# - On step 16, the candidate index 30 is considered, with corresponding
#   coefficient :math:`a_{30} = -1.612`. Since this coefficient has an absolute
#   value greater than the threshold, it gets selected and the active multi-index
#   set is [0, 1, 7, 10, 15, 30] on the end of this step. From this step to the
#   end, the index 30 will not leave the active set.
# - On the step 20, the index 35 enters the active set.
# - On the step 25, the index 40 enters the active set.
# - On the step 34, the index 49 enters the active set.
# - On the step 69, the index 84 enters the active set.
# - On the step 74, the index 89 enters the active set.
# - On the step 83, the index 98 enters the active set.
# - On the last step, the active multi-index set contains the indices
#   [0, 1, 7, 10, 15, 30, 35, 40, 49, 84, 89, 98] and the corresponding
#   coefficients are [3.508, 1.625, -0.6414, -1.291, 1.372, -1.613, 0.2076,
#   -1.090, 0.4092, -0.2078, 0.1753, -0.3250].
#
# We see that the algorithm was able so select 12 coefficients in the first
# 101 coefficients considered by the algorithm. It could have selected
# more coefficients since we provided 16 slots to fill thanks to the
# `mostSignificant` parameter. The considered coefficients were, however,
# too close to zero and were below the threshold.


# %%
#
# Conclusion
# ----------
#
# We see that the :class:`~openturns.CleaningStrategy` class performs correctly in
# this particular case. We have seen how to select the hyperparameters which
# produce the best :math:`R^2` score.

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Create a function
inputVar = ot.Description(2)
inputVar[0] = "x"
inputVar[1] = "y"
formula = ot.Description(1)
formula[0] = "exp(-sin(cos(y)^2*x^2+sin(x)^2*y^2))"
f = ot.SymbolicFunction(inputVar, formula)

# Generate the data for the curves to be drawn
discretization = ot.Point(2)
nX = 75
nY = 75
discretization[0] = nX
discretization[1] = nY
inputData = ot.Box(discretization).generate()
inputData *= [10.0] * 2
inputData += [-5.0] * 2
data = f(inputData)
size = 7
levels = ot.Point(size)
for i in range(size):
    levels[i] = (0.5 + i) / size
# Create an empty graph
myGraph = ot.Graph("Complex iso lines", "u1", "u2")
myGraph.setLegendPosition("upper right")

# Create the first cloud
myContour = ot.Contour(nX + 2, nY + 2, data)
myContour.setLevels(levels)
myContour.setColor("red")
print("contour=", myContour)

# Modify it to filled Contour
myContour.setColorMap("inferno")
myContour.setAlpha(0.5)
myContour.setExtend("min")
myContour.setColorBarPosition("left")
myContour.setColorMapNorm("symlog")
myContour.setVmin(3)
myContour.setHatches(ot.Description(["/", "\\", "+/", "*"]))
myContour.setIsFilled(True)
print("filled contour=", myContour.__repr__())

# Then, draw it
myGraph.add(myContour)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott
import sys

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(3)
ot.Log.Show(ot.Log.ALL)


def progress(percent):
    sys.stderr.write("-- progress=" + str(percent) + "%\n")


def stop():
    sys.stderr.write("-- stop?\n")
    return False


algoNames = ot.Ceres.GetAlgorithmNames()
print(algoNames)

# unconstrained optimization: rosenbrock x*=(1,1), x*=(0.7864, 0.6177) on unit disk
dim = 2
f = ot.SymbolicFunction(["x1", "x2"], ["1+100*(x2-x1^2)^2+(1-x1)^2"])
startingPoint = [1e-3] * dim
p_ref = [1.0] * dim

for algoName in algoNames:
    if algoName in ["LEVENBERG_MARQUARDT", "DOGLEG"]:
        # only test general optimization algorithms
        continue
    for minimization in [True, False]:
        if algoName == "NONLINEAR_CONJUGATE_GRADIENT" and not minimization:
            # goes very far and the function cannot evaluate
            continue
        print("algoName=", algoName, "minimization=", minimization)
        problem = ot.OptimizationProblem(f)
        problem.setMinimization(minimization)
        algo = ot.Ceres(problem, algoName)
        algo.setMaximumIterationNumber(100000)
        algo.setStartingPoint(startingPoint)
        # algo.setProgressCallback(progress)
        # algo.setStopCallback(stop)
        algo.run()
        result = algo.getResult()
        x_star = result.getOptimalPoint()
        if minimization and algoName != "STEEPEST_DESCENT":
            ott.assert_almost_equal(x_star, p_ref, 5e-2)
        print(result)


# least-squares optimization
n = 3
m = 10

x = [[0.5 + i] for i in range(m)]


model = ot.SymbolicFunction(["a", "b", "c", "x"], ["a + b * exp(min(500, c * x))"])
p_ref = [2.8, 1.2, 0.5]  # a, b, c
modelx = ot.ParametricFunction(model, [0, 1, 2], p_ref)
y = modelx(x)


def residualFunction_py(p):
    modelx = ot.ParametricFunction(model, [0, 1, 2], p)
    return [modelx(x[i])[0] - y[i, 0] for i in range(m)]


residualFunction = ot.PythonFunction(n, m, residualFunction_py)

bounds = ot.Interval([0, 0, 0], [2.5, 8.0, 19])

for algoName in algoNames:
    line_search = not (algoName in ["LEVENBERG_MARQUARDT", "DOGLEG"])
    for bound in [True, False]:
        if bound and line_search:
            # line search do not support bound constraints
            continue
        print("algoName=", algoName, "bound=", bound)
        problem = ot.LeastSquaresProblem(residualFunction)
        if bound:
            problem.setBounds(bounds)
        startingPoint = [1.0] * n
        algo = ot.Ceres(problem, algoName)
        algo.setStartingPoint(startingPoint)
        # algo.setProgressCallback(progress)
        # algo.setStopCallback(stop)
        algo.setMaximumIterationNumber(10000)
        algo.run()
        result = algo.getResult()
        x_star = result.getOptimalPoint()
        print(result)
        if bound:
            assert x_star in bounds, "optimal point not in bounds"
        else:
            if not line_search:
                # line search algorithms converge less well
                ott.assert_almost_equal(x_star, p_ref, 0.1)



"""
Create a discrete Markov chain process
======================================
"""

# %%
# This example details first how to create and manipulate a discrete Markov chain.
#
# A discrete Markov chain :math:`X: \Omega \times \mathcal{D} \rightarrow E`, where :math:`E = [\![ 0,...,p-1]\!]` is a process
# where :math:`\mathcal{D}=\mathbb{R}` discretized on the time grid :math:`(t_k)_{k \geq 0}` such
# that:
#
# .. math::
#    \begin{aligned}
#      \forall k > 0,\: \mathbb{P} ( X_{t_k} \> | \> X_{t_0},...X_{t_{k-1}} )  =  \mathbb{P} ( X_{t_k} \> | \> X_{t_{k-1}} )
#   \end{aligned}
#
#
# The transition matrix of the process :math:`\mathcal{M} = (m_{i,j})` can be defined such that:
#
# .. math::
#     \begin{aligned}
#         \forall t_k \in \mathcal{D}, \forall i,j < p , \> m_{i+1,j+1} = \mathbb{P} (X_{t_{k+1}} = j \> | \> X_{t_{k}} = i)
#     \end{aligned}
#
# The library proposes to model it through the object :class:`~openturns.DiscreteMarkovChain` defined thanks to the origin :math:`X_{t_0}`
# (which can be either deterministic or uncertain), the transition matrix :math:`\mathcal{M}` and the time grid.

# %%
import openturns as ot
import openturns.viewer as otv
from matplotlib import pyplot as plt

# %%
# Define a generic function to plot matrices


def plotMatrix(matrix, texts=False, origin=None, colorbar=False, extent=None, **kwargs):
    """Generic procedure for displaying a matrix with or without text overlay and color bar"""
    res = plt.matshow(matrix, origin=origin, extent=extent, **kwargs)
    if texts:
        if extent is None:
            extent = (-0.5, matrix.getNbColumns() - 0.5, -0.5, matrix.getNbRows() - 0.5)
        x_step = (extent[1] - extent[0]) / matrix.getNbColumns()
        y_step = (extent[3] - extent[2]) / matrix.getNbRows()
        for i in range(matrix.getNbColumns()):
            for j in range(matrix.getNbRows()):
                c = round(
                    matrix[j if origin == "lower" else (matrix.getNbRows() - j - 1), i],
                    2,
                )
                plt.text(
                    i * x_step + extent[0] + x_step / 2,
                    j * y_step + extent[2] + y_step / 2,
                    str(c),
                    va="center",
                    ha="center",
                )
    if colorbar:
        plt.colorbar(res)


# %%
# Define the origin
origin = ot.Dirac(0.0)

# %%
# Define the transition matrix
transition = ot.SquareMatrix(
    [
        [0.1, 0.3, 0.5, 0.1],
        [0.6, 0.1, 0.2, 0.1],
        [0.4, 0.3, 0.1, 0.2],
        [0.2, 0.0, 0.1, 0.7],
    ]
)

# %%
# Visualize the transition matrix
_ = plt.matshow(transition)

# %%
# Invert axes and add texts
plotMatrix(
    transition,
    cmap="gray",
    texts=True,
    origin="lower",
    colorbar=True,
    alpha=0.5,
    vmin=0,
    vmax=1,
)

# %%
# Define an 1-d mesh
tgrid = ot.RegularGrid(0.0, 1.0, 50)

# %%
# Markov chain definition and realization
process = ot.DiscreteMarkovChain(origin, transition, tgrid)
real = process.getRealization()
graph = real.drawMarginal(0)
graph.setTitle("Discrete Markov chain")
view = otv.View(graph)

# %%
# Get several realizations
process.setTimeGrid(ot.RegularGrid(0.0, 1.0, 20))
reals = process.getSample(3)
graph = reals.drawMarginal(0)
graph.setTitle("Discrete Markov chain, 3 realizations")
view = otv.View(graph)

# %%
# Markov chain future 10 steps
future = process.getFuture(10)
graph = future.drawMarginal(0)
graph.setTitle("Markov chain future 10 steps")
view = otv.View(graph)

# %%
# Markov chain 3 different futures
futures = process.getFuture(10, 3)
graph = futures.drawMarginal(0)
graph.setTitle("Three Markov chain futures, 10 steps")
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Use case : Branin test function
===============================
"""

import openturns as ot


class BraninModel:
    r"""
    Data class for the Branin test function.

    Attributes
    ----------

    dim : int
        The dimension of the problem, dim=2.

    model : :class:`~openturns.Function`
        The Branin function

    trueNoiseFunction : float
        Constant, small noise :math:`\epsilon=0.1`.

    noiseModel : :class:`~openturns.Function`
        The noise function

    lowerbound : :class:`~openturns.Point` in dimension dim.
        Default is 0.0 for each dimension.

    upperbound : :class:`~openturns.Point` in dimension dim.
        Default is 1.0 for each dimension.

    xexact1 : :class:`~openturns.Point`
        First minima location :math:`[0.123895, 0.818329]`.

    xexact2 : :class:`~openturns.Point`
        Second minima location :math:`[0.542773, 0.151666]`.

    xexact3 : :class:`~openturns.Point`
        Third minima location :math:`[0.961652, 0.165000]`.

    xexact : :class:`~openturns.Sample`
        All global minimas location gathered.

    Examples
    --------
    >>> from openturns.usecases import branin_function
    >>> # Load the Branin model
    >>> bm = branin_function.BraninModel()
    """

    def __init__(self):
        self.dim = 2
        self.trueNoiseFunction = 0.1

        self.branin = ot.SymbolicFunction(
            ["x1", "x2"],
            [
                "((x2-(5.1/(4*pi_^2))*x1^2+5*x1/pi_-6)^2+10*(1-1/(8*pi_))*cos(x1)+10-54.8104)/51.9496"
            ],
        )
        self.transfo = ot.SymbolicFunction(["u1", "u2"], ["15*u1-5", "15*u2"])

        self.model = ot.ComposedFunction(self.branin, self.transfo)
        self.noiseModel = ot.SymbolicFunction(
            ["x1", "x2"], [str(self.trueNoiseFunction)]
        )
        # %%
        self.lowerbound = ot.Point([0.0] * self.dim)
        self.upperbound = ot.Point([1.0] * self.dim)

        self.objectiveFunction = self.model

        # Minimas
        self.xexact1 = ot.Point([0.123895, 0.818329])
        self.xexact2 = ot.Point([0.542773, 0.151666])
        self.xexact3 = ot.Point([0.961652, 0.165000])
        self.xexact = ot.Sample([self.xexact1, self.xexact2, self.xexact3])



"""
Kolmogorov-Smirnov : understand the p-value
===========================================
"""

# %%
# In this example, we illustrate the calculation of the Kolmogorov-Smirnov (KS) p-value.
#
# * We generate a sample from a Gaussian distribution.
# * We create a uniform distribution with known parameters.
# * The Kolmogorov-Smirnov statistics is computed and plot on the empirical cumulative distribution function.
# * We plot the p-value as the area under the part of the curve exceeding the observed statistics.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# We generate a sample from a standard Gaussian distribution.
dist = ot.Normal()
samplesize = 10
sample = dist.getSample(samplesize)

# %%
testdistribution = ot.Normal()
result = ot.FittingTest.Kolmogorov(sample, testdistribution, 0.01)

# %%
pvalue = result.getPValue()
pvalue

# %%
KSstat = result.getStatistic()
KSstat

# %%
# Compute exact Kolmogorov PDF.

# %%
# Create a function which returns the CDF given the KS distance.


# %%
def pKolmogorovPy(x):
    y = ot.DistFunc.pKolmogorov(samplesize, x[0])
    return [y]


# %%
pKolmogorov = ot.PythonFunction(1, 1, pKolmogorovPy)


# %%
# Create a function which returns the KS PDF given the KS distance: use the `gradient` method.


# %%
def kolmogorovPDF(x):
    return pKolmogorov.gradient(x)[0, 0]


# %%
def dKolmogorov(x):
    """
    Compute Kolmogorov PDF for given x.
    x : a Sample, the points where the PDF must be evaluated
    Reference
    Numerical Derivatives in Scilab, Michael Baudin, May 2009
    """
    n = x.getSize()
    y = ot.Sample(n, 1)
    for i in range(n):
        y[i, 0] = kolmogorovPDF(x[i])
    return y


# %%
def linearSample(xmin, xmax, npoints):
    """Returns a sample created from a regular grid
    from xmin to xmax with npoints points."""
    step = (xmax - xmin) / (npoints - 1)
    rg = ot.RegularGrid(xmin, step, npoints)
    vertices = rg.getVertices()
    return vertices


# %%
n = 1000  # Number of points in the plot
s = linearSample(0.001, 0.999, n)
y = dKolmogorov(s)


# %%
# Create a regular grid starting from the observed KS statistics.

# %%
nplot = 100
x = linearSample(KSstat, 0.6, nplot)

# %%
# Compute the bounds to fill: the lower vertical bound is 0 and the upper vertical bound is the KS PDF.

# %%
vLow = [0.0] * nplot
vUp = [pKolmogorov.gradient(x[i])[0, 0] for i in range(nplot)]

# %%
boundsPoly = ot.Polygon.FillBetween(x.asPoint(), vLow, vUp)
boundsPoly.setLegend("pvalue = %.4f" % (pvalue))
curve = ot.Curve(s, y)
curve.setLegend("Exact distribution")
curveStat = ot.Curve([KSstat, KSstat], [0.0, kolmogorovPDF([KSstat])])
curveStat.setColor("red")
curveStat.setLegend("KS-statistics = %.4f" % (KSstat))
graph = ot.Graph(
    "Kolmogorov-Smirnov distribution (known parameters)",
    "KS-Statistics",
    "PDF",
    True,
    "upper right",
)
graph.setLegends(["Empirical distribution"])
graph.add(curve)
graph.add(curveStat)
graph.add(boundsPoly)
graph.setTitle("Kolmogorov-Smirnov distribution (known parameters)")
view = otv.View(graph)

# %%
# We observe that the p-value is the area of the curve which corresponds to
# the KS distances greater than the observed KS statistics.

# %%
otv.View.ShowAll()



import openturns as ot
import openturns.viewer as otv


# 3-d test
R1 = ot.CovarianceMatrix(3)
R1[2, 1] = -0.25
R2 = ot.CovarianceMatrix(3)
R2[1, 0] = 0.5
R2[2, 1] = -0.3
R2[0, 0] = 1.3
print(R2)
dists = [ot.Normal([1.0, -2.0, 3.0], R1), ot.Normal([-1.0, 2.0, -2.0], R2)]
mixture = ot.Mixture(dists, [2.0 / 3.0, 1.0 / 3.0])

# 2-d test
dists = [
    ot.Normal([-1.0, 2.0], [1.0] * 2, ot.CorrelationMatrix(2)),
    ot.Normal([1.0, -2.0], [1.5] * 2, ot.CorrelationMatrix(2)),
]
mixture = ot.Mixture(dists)

sample = mixture.getSample(100)
distribution = ot.KernelSmoothing().build(sample)
algo = ot.MinimumVolumeClassifier(distribution, [0.8])
graph = algo.drawContourAndSample([0.1, 0.5, 0.8], sample, [0, 1])

otv.View(graph, contour_kw={"colors": ["black"]}, figure_kw={"figsize": (6.0, 6.0)})



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Polya(15, 0.7)
size = 10000
sample = distribution.getSample(size)
factory = ot.PolyaFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedPolya = factory.buildAsPolya(sample)
print("Polya          =", distribution)
print("Estimated polya=", estimatedPolya)
estimatedPolya = factory.buildAsPolya()
print("Default polya=", estimatedPolya)
estimatedPolya = factory.buildAsPolya(distribution.getParameter())
print("Polya from parameters=", estimatedPolya)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

ot.ResourceMap.SetAsUnsignedInteger("MeixnerDistribution-CDFDiscretization", 100)
ot.ResourceMap.SetAsScalar("MeixnerDistribution-MaximumAbsoluteError", 1.0e-6)
ot.ResourceMap.SetAsScalar("MeixnerDistribution-MaximumRelativeError", 1.0e-6)
ot.ResourceMap.SetAsScalar("MeixnerDistribution-MaximumConstraintError", 1.0e-6)
ot.ResourceMap.SetAsScalar("MeixnerDistribution-MaximumObjectiveError", 1.0e-6)
distribution = ot.MeixnerDistribution(1.5, 0.5, 2.5, -0.5)
size = 1000
sample = distribution.getSample(size)
factory = ot.MeixnerDistributionFactory()
covariance = ot.CovarianceMatrix()
# Distribution estimatedDistribution(factory.build(sample, covariance))
estimatedDistribution = factory.build(sample)
print("Distribution          =", distribution)
print("Estimated distribution=", estimatedDistribution)
# print "Covariance=", covariance
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedMeixnerDistribution = factory.buildAsMeixnerDistribution(sample)
print("MeixnerDistribution          =", distribution)
print("Estimated meixnerDistribution=", estimatedMeixnerDistribution)
estimatedMeixnerDistribution = factory.buildAsMeixnerDistribution()
print("Default meixnerDistribution=", estimatedMeixnerDistribution)
estimatedMeixnerDistribution = factory.buildAsMeixnerDistribution(
    distribution.getParameter()
)
print("MeixnerDistribution from parameters=", estimatedMeixnerDistribution)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

# Default dimension parameter to evaluate the model
inputDimension = 1
outputDimension = 1

# Amplitude values
amplitude = [1.0] * outputDimension
# Scale values
scale = [1.0] * inputDimension

tmin = 0.0
step = 0.1
n = 11

myTimeGrid = ot.RegularGrid(tmin, step, n)
size = 25

# Second order model with parameters
myCovModel = ot.ExponentialModel(scale, amplitude)
print("myCovModel = ", myCovModel)

myProcess1 = ot.GaussianProcess(myCovModel, myTimeGrid)
print("myProcess1 = ", myProcess1)
print("is stationary? ", myProcess1.isStationary())
myProcess1.setSamplingMethod(ot.GaussianProcess.CHOLESKY)
print("mean over ", size, " realizations = ", myProcess1.getSample(size).computeMean())
myProcess1.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
print("mean over ", size, " realizations = ", myProcess1.getSample(size).computeMean())

# With constant trend
trend = ot.TrendTransform(ot.SymbolicFunction("t", "4.0"), myTimeGrid)
myProcess2 = ot.GaussianProcess(trend, myCovModel, myTimeGrid)
myProcess2.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
print("myProcess2 = ", myProcess2)
print("is stationary? ", myProcess2.isStationary())
print("mean over ", size, " realizations= ", myProcess2.getSample(size).computeMean())

# With varying trend
trend3 = ot.TrendTransform(ot.SymbolicFunction("t", "sin(t)"), myTimeGrid)
myProcess3 = ot.GaussianProcess(trend3, myCovModel, myTimeGrid)
print("myProcess3 = ", myProcess3)
print("is stationary? ", myProcess3.isStationary())
myProcess3.setSamplingMethod(ot.GaussianProcess.CHOLESKY)
print("mean over ", size, " realizations = ", myProcess3.getSample(size).computeMean())
myProcess3.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
print("mean over ", size, " realizations = ", myProcess3.getSample(size).computeMean())

cov = ot.ExponentialModel([2.0] * inputDimension, [1.0, 2.0, 3.0])
model = ot.GaussianProcess(cov, myTimeGrid)
print("model=", model)
print("marginal=", model.getMarginal([0, 2]))

# FIX #2121
ot.RandomGenerator.SetSeed(0)
standard_deviation = 10.0
mesh = ot.Mesh([[0.0]])  # singleton
cov_matrix = ot.CovarianceMatrix(
    1, [standard_deviation**2]
)  # associated "covariance matrix"
covModel = ot.UserDefinedCovarianceModel(mesh, cov_matrix)

# Create the "Gaussian Process" discretized on a singleton
myProcess = ot.GaussianProcess(covModel, mesh)
myProcess.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
size = 1000
sample = ot.Sample(size, 1)
for i in range(size):
    sample[i] = myProcess.getRealization().getValues()[0]

ott.assert_almost_equal(sample.computeStandardDeviation()[0], 10.0676)

# Multivariate outputs
ot.RandomGenerator.SetSeed(1)
inputDimension = 1
outputDimension = 2

size = 1000
ot.ResourceMap.SetAsUnsignedInteger("GaussianProcess-GibbsMaximumIteration", 1000)
# Amplitude values
amplitude = [2.0, 3.0]
# Scale values
scale = [1.0] * inputDimension

myCovModel = ot.ExponentialModel(scale, amplitude)

myProcess4 = ot.GaussianProcess(myCovModel, myTimeGrid)
print("myProcess4 = ", myProcess4)
print("is stationary? ", myProcess4.isStationary())
myProcess4.setSamplingMethod(ot.GaussianProcess.CHOLESKY)
sample = myProcess4.getSample(size)
print("mean over ", size, " realizations = ", sample.computeMean())
print("variance over ", size, " realizations = ", sample.computeVariance())
myProcess4.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
sample = myProcess4.getSample(size)
print("mean over ", size, " realizations = ", sample.computeMean())
print("variance over ", size, " realizations = ", sample.computeVariance())

trend5 = ot.TrendTransform(
    ot.SymbolicFunction(["t"], ["sin(pi_ / 2 * t)", "2 * sin(pi_ / 2 * t)"]), myTimeGrid
)
myProcess5 = ot.GaussianProcess(trend5, myCovModel, myTimeGrid)
print("myProcess5 = ", myProcess5)
print("is stationary? ", myProcess5.isStationary())
myProcess5.setSamplingMethod(ot.GaussianProcess.CHOLESKY)
sample = myProcess5.getSample(size)
print("mean over ", size, " realizations = ", sample.computeMean())
print("variance over ", size, " realizations = ", sample.computeVariance())
myProcess5.setSamplingMethod(ot.GaussianProcess.GALLIGAOGIBBS)
sample = myProcess5.getSample(size)
print("mean over ", size, " realizations = ", sample.computeMean())
print("variance over ", size, " realizations = ", sample.computeVariance())
